@@ -2,8 +2,8 @@
 
   dir.c -
 
-  $Author: eban $
-  $Date: 2005/04/11 13:26:32 $
+  $Author: matz $
+  $Date: 2006/05/17 09:14:34 $
   created at: Wed Jan  5 09:51:01 JST 1994
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -325,7 +325,17 @@ dir_closed()
     rb_raise(rb_eIOError, "closed directory");
 }
 
+static void
+dir_check(dir)
+    VALUE dir;
+{
+    if (!OBJ_TAINTED(dir) && rb_safe_level() >= 4)
+	rb_raise(rb_eSecurityError, "Insecure: operation on untainted Dir");
+    rb_check_frozen(dir);
+}
+
 #define GetDIR(obj, dirp) do {\
+    dir_check(dir);\
     Data_Get_Struct(obj, struct dir_data, dirp);\
     if (dirp->dir == NULL) dir_closed();\
 } while (0)
@@ -409,6 +419,7 @@ dir_each(dir)
     struct dirent *dp;
 
     GetDIR(dir, dirp);
+    rewinddir(dirp->dir);
     for (dp = readdir(dirp->dir); dp != NULL; dp = readdir(dirp->dir)) {
 	rb_yield(rb_tainted_str_new(dp->d_name, NAMLEN(dp)));
 	if (dirp->dir == NULL) dir_closed();
@@ -535,6 +546,9 @@ dir_close(dir)
 {
     struct dir_data *dirp;
 
+    if (rb_safe_level() >= 4 && !OBJ_TAINTED(dir)) {
+	rb_raise(rb_eSecurityError, "Insecure: can't close");
+    }
     GetDIR(dir, dirp);
     closedir(dirp->dir);
     dirp->dir = NULL;
@@ -785,6 +799,17 @@ dir_s_rmdir(obj, dir)
     return INT2FIX(0);
 }
 
+static void
+sys_warning_1(mesg)
+    const char* mesg;
+{
+    rb_sys_warning("%s", mesg);
+}
+
+#define GLOB_VERBOSE	(1 << (sizeof(int) * CHAR_BIT - 1))
+#define sys_warning(val) \
+    ((flags & GLOB_VERBOSE) && rb_protect((VALUE (*)_((VALUE)))sys_warning_1, (VALUE)(val), 0))
+
 /* Return nonzero if S has any special globbing chars in it.  */
 static int
 has_magic(s, send, flags)
@@ -877,8 +902,8 @@ remove_backslashes(p)
 #endif
 
 struct glob_args {
-    void (*func) _((VALUE, VALUE));
-    VALUE c;
+    void (*func) _((const char*, VALUE));
+    const char *c;
     VALUE v;
 };
 
@@ -889,64 +914,56 @@ glob_func_caller(val)
     VALUE val;
 {
     struct glob_args *args = (struct glob_args *)val;
-    VALUE path = args->c;
 
-    OBJ_TAINT(path);
-    RSTRING(path)->len = strlen(RSTRING(path)->ptr);
-    (*args->func)(path, args->v);
+    (*args->func)(args->c, args->v);
     return Qnil;
 }
 
-static int
-glob_call_func(func, path, arg)
-    void (*func) _((VALUE, VALUE));
-    VALUE path;
-    VALUE arg;
-{
-    int status;
-    struct glob_args args;
+#define glob_call_func(func, path, arg) (*func)(path, arg)
 
-    args.func = func;
-    args.c = path;
-    args.v = arg;
-
-    rb_protect(glob_func_caller, (VALUE)&args, &status);
-    return status;
-}
-
-static int glob_helper(VALUE path, char *sub, int flags, void (*func)(VALUE,VALUE), VALUE arg);
+static int glob_helper _((const char *path, const char *sub, int flags, int (*func)(const char *,VALUE), VALUE arg));
 
 static int
-glob_helper(pv, sub, flags, func, arg)
-    VALUE pv;
-    char *sub;
+glob_helper(path, sub, flags, func, arg)
+    const char *path;
+    const char *sub;
     int flags;
-    void (*func) _((VALUE, VALUE));
+    int (*func) _((const char *, VALUE));
     VALUE arg;
 {
     struct stat st;
-    char *p, *m, *path;
+    const char *p, *m;
     int status = 0;
+    char *buf = 0;
+    char *newpath = 0;
 
-    StringValue(pv);
-    path = RSTRING(pv)->ptr;
     p = sub ? sub : path;
     if (!has_magic(p, 0, flags)) {
-#if defined DOSISH
-	remove_backslashes(path);
-#else
-	if (!(flags & FNM_NOESCAPE)) remove_backslashes(p);
+#if !defined DOSISH
+	if (!(flags & FNM_NOESCAPE))
 #endif
+	{
+	    newpath = strdup(path);
+	    if (sub) {
+		p = newpath + (sub - path);
+		remove_backslashes(newpath + (sub - path));
+		sub = p;
+	    }
+	    else {
+		remove_backslashes(newpath);
+		p = path = newpath;
+	    }
+	}
 	if (lstat(path, &st) == 0) {
-	    status = glob_call_func(func, pv, arg);
-	    if (status) return status;
+	    status = glob_call_func(func, path, arg);
 	}
 	else if (errno != ENOENT) {
 	    /* In case stat error is other than ENOENT and
 	       we may want to know what is wrong. */
-	    rb_sys_warning(path);
+	    sys_warning(path);
 	}
-	return 0;
+	xfree(newpath);
+	return status;
     }
 
     while (p && !status) {
@@ -954,7 +971,6 @@ glob_helper(pv, sub, flags, func, arg)
 	m = strchr(p, '/');
 	if (has_magic(p, m, flags)) {
 	    char *dir, *base, *magic;
-	    VALUE buf;
 	    DIR *dirp;
 	    struct dirent *dp;
 	    int recursive = 0;
@@ -970,7 +986,8 @@ glob_helper(pv, sub, flags, func, arg)
 
 	    magic = extract_elem(p);
 	    if (stat(dir, &st) < 0) {
-	        if (errno != ENOENT) rb_sys_warning(dir);
+	        if (errno != ENOENT)
+		    sys_warning(dir);
 	        free(base);
 	        free(magic);
 	        break;
@@ -979,14 +996,14 @@ glob_helper(pv, sub, flags, func, arg)
 		if (m && strcmp(magic, "**") == 0) {
 		    int n = strlen(base);
 		    recursive = 1;
-		    buf = rb_str_new(0, n+strlen(m)+3);
-		    sprintf(RSTRING(buf)->ptr, "%s%s", base, *base ? m : m+1);
-		    status = glob_helper(buf, RSTRING(buf)->ptr+n, flags, func, arg);
+		    REALLOC_N(buf, char, n+strlen(m)+3);
+		    sprintf(buf, "%s%s", base, *base ? m : m+1);
+		    status = glob_helper(buf, buf+n, flags, func, arg);
 		    if (status) goto finalize;
 		}
 		dirp = opendir(dir);
 		if (dirp == NULL) {
-		    rb_sys_warning(dir);
+		    sys_warning(dir);
 		    free(base);
 		    free(magic);
 		    break;
@@ -1010,14 +1027,15 @@ glob_helper(pv, sub, flags, func, arg)
 			continue;
 		    if (fnmatch("*", dp->d_name, flags) != 0)
 			continue;
-		    buf = rb_str_new(0, strlen(base)+NAMLEN(dp)+strlen(m)+6);
-		    sprintf(RSTRING(buf)->ptr, "%s%s%s", base, (BASE) ? "/" : "", dp->d_name);
-		    if (lstat(RSTRING(buf)->ptr, &st) < 0) {
-			if (errno != ENOENT) rb_sys_warning(RSTRING(buf)->ptr);
+		    REALLOC_N(buf, char, strlen(base)+NAMLEN(dp)+strlen(m)+6);
+		    sprintf(buf, "%s%s%s", base, (BASE) ? "/" : "", dp->d_name);
+		    if (lstat(buf, &st) < 0) {
+			if (errno != ENOENT)
+			    sys_warning(buf);
 			continue;
 		    }
 		    if (S_ISDIR(st.st_mode)) {
-			char *t = RSTRING(buf)->ptr+strlen(RSTRING(buf)->ptr);
+			char *t = buf+strlen(buf);
 		        strcpy(t, "/**");
 			strcpy(t+3, m);
 			status = glob_helper(buf, t, flags, func, arg);
@@ -1027,15 +1045,16 @@ glob_helper(pv, sub, flags, func, arg)
 		    continue;
 		}
 		if (fnmatch(magic, dp->d_name, flags) == 0) {
-		    buf = rb_str_new(0, strlen(base)+NAMLEN(dp)+2);
-		    sprintf(RSTRING(buf)->ptr, "%s%s%s", base, (BASE) ? "/" : "", dp->d_name);
+		    REALLOC_N(buf, char, strlen(base)+NAMLEN(dp)+2);
+		    sprintf(buf, "%s%s%s", base, (BASE) ? "/" : "", dp->d_name);
 		    if (!m) {
 			status = glob_call_func(func, buf, arg);
 			if (status) break;
 			continue;
 		    }
 		    tmp = ALLOC(struct d_link);
-		    tmp->path = strdup(RSTRING(buf)->ptr);
+		    tmp->path = buf;
+		    buf = 0;
 		    *tail = tmp;
 		    tail = &tmp->next;
 		}
@@ -1053,13 +1072,13 @@ glob_helper(pv, sub, flags, func, arg)
 				int len = strlen(link->path);
 				int mlen = strlen(m);
 
-				buf = rb_str_new(0, len+mlen+1);
-				sprintf(RSTRING(buf)->ptr, "%s%s", link->path, m);
-				status = glob_helper(buf, RSTRING(buf)->ptr+len, flags, func, arg);
+				REALLOC_N(buf, char, len+mlen+1);
+				sprintf(buf, "%s%s", link->path, m);
+				status = glob_helper(buf, buf+len, flags, func, arg);
 			    }
 			}
 			else {
-			    rb_sys_warning(link->path);
+			    sys_warning(link->path);
 			}
 		    }
 		    tmp = link;
@@ -1072,33 +1091,59 @@ glob_helper(pv, sub, flags, func, arg)
 	}
 	p = m;
     }
+    xfree(buf);
+    xfree(newpath);
+    return status;
+}
+
+int
+ruby_glob(path, flags, func, arg)
+    const char *path;
+    int flags;
+    int (*func) _((const char *, VALUE));
+    VALUE arg;
+{
+    return glob_helper(path, 0, flags & ~GLOB_VERBOSE, func, arg);
+}
+
+int
+ruby_globi(path, flags, func, arg)
+    const char *path;
+    int flags;
+    int (*func) _((const char *, VALUE));
+    VALUE arg;
+{
+    return glob_helper(path, 0, flags | FNM_CASEFOLD, func, arg);
+}
+
+static int rb_glob_caller _((const char *, VALUE));
+
+static int
+rb_glob_caller(path, a)
+    const char *path;
+    VALUE a;
+{
+    int status;
+    struct glob_args *args = (struct glob_args *)a;
+
+    args->c = path;
+    rb_protect(glob_func_caller, a, &status);
     return status;
 }
 
 static int
 rb_glob2(path, flags, func, arg)
-    VALUE path;
+    const char *path;
     int flags;
-    void (*func) _((VALUE, VALUE));
+    void (*func) _((const char *, VALUE));
     VALUE arg;
 {
-    int status;
+    struct glob_args args;
 
-    status = glob_helper(path, 0, flags, func, arg);
-    return status;
-}
+    args.func = func;
+    args.v = arg;
 
-struct rb_glob_args {
-    void (*func) _((const char*, VALUE));
-    VALUE arg;
-};
-
-static void
-rb_glob_caller(path, a)
-    VALUE path, a;
-{
-    struct rb_glob_args *args = (struct rb_glob_args *)a;
-    (*args->func)(RSTRING(path)->ptr, args->arg);
+    return glob_helper(path, 0, flags | GLOB_VERBOSE, rb_glob_caller, (VALUE)&args);
 }
 
 void
@@ -1107,13 +1152,7 @@ rb_glob(path, func, arg)
     void (*func) _((const char*, VALUE));
     VALUE arg;
 {
-    struct rb_glob_args args;
-    int status;
-
-    args.func = func;
-    args.arg = arg;
-    status = rb_glob2(rb_str_new2(path), 0, rb_glob_caller, (VALUE)&args);
-
+    int status = rb_glob2(path, 0, func, arg);
     if (status) rb_jump_tag(status);
 }
 
@@ -1123,28 +1162,22 @@ rb_globi(path, func, arg)
     void (*func) _((const char*, VALUE));
     VALUE arg;
 {
-    struct rb_glob_args args;
-    int status;
-
-    args.func = func;
-    args.arg = arg;
-    status = rb_glob2(rb_str_new2(path), FNM_CASEFOLD, rb_glob_caller, (VALUE)&args);
-
+    int status = rb_glob2(path, FNM_CASEFOLD, func, arg);
     if (status) rb_jump_tag(status);
 }
 
 static void
 push_pattern(path, ary)
-    VALUE path;
+    const char *path;
     VALUE ary;
 {
-    rb_ary_push(ary, path);
+    rb_ary_push(ary, rb_tainted_str_new2(path));
 }
 
 static int
 push_globs(ary, s, flags)
     VALUE ary;
-    VALUE s;
+    const char *s;
     int flags;
 {
     return rb_glob2(s, flags, push_pattern, ary);
@@ -1152,17 +1185,18 @@ push_globs(ary, s, flags)
 
 static int
 push_braces(ary, str, flags)
-    VALUE ary, str;
+    VALUE ary;
+    const char *str;
     int flags;
 {
-    VALUE buf;
+    char *buf = 0;
     char *b;
     const char *s, *p, *t;
     const char *lbrace, *rbrace;
     int nest = 0;
     int status = 0;
 
-    s = p = RSTRING(str)->ptr;
+    s = p = str;
     lbrace = rbrace = 0;
     while (*p) {
 	if (*p == '{') {
@@ -1187,11 +1221,16 @@ push_braces(ary, str, flags)
 	    t = p + 1;
 	    for (p = t; *p!='}' && *p!=','; p++) {
 		/* skip inner braces */
-		if (*p == '{') while (*p!='}') p++;
+		if (*p == '{') {
+		    nest = 1;
+		    while (*++p != '}' || --nest) {
+			if (*p == '{') nest++;
+		    }
+		}
 	    }
-	    buf = rb_str_new(0, len+1);
-	    memcpy(RSTRING(buf)->ptr, s, lbrace-s);
-	    b = RSTRING(buf)->ptr + (lbrace-s);
+	    REALLOC_N(buf, char, len+1);
+	    memcpy(buf, s, lbrace-s);
+	    b = buf + (lbrace-s);
 	    memcpy(b, t, p-t);
 	    strcpy(b+(p-t), rbrace+1);
 	    status = push_braces(ary, buf, flags);
@@ -1201,6 +1240,7 @@ push_braces(ary, str, flags)
     else {
 	status = push_globs(ary, str, flags);
     }
+    xfree(buf);
 
     return status;
 }
@@ -1212,9 +1252,7 @@ rb_push_glob(str, flags)
     VALUE str;
     int flags;
 {
-    const char *p, *pend;
-    volatile VALUE buf;
-    char *t;
+    const char *p, *pend, *buf;
     int nest, maxnest;
     int status = 0;
     int noescape = flags & FNM_NOESCAPE;
@@ -1226,20 +1264,17 @@ rb_push_glob(str, flags)
     pend = p + RSTRING(str)->len;
 
     while (p < pend) {
-	buf = rb_str_new(0, pend - p + 1);
-	t = RSTRING(buf)->ptr;
 	nest = maxnest = 0;
 	while (p < pend && isdelim(*p)) p++;
+	buf = p;
 	while (p < pend && !isdelim(*p)) {
 	    if (*p == '{') nest++, maxnest++;
 	    if (*p == '}') nest--;
 	    if (!noescape && *p == '\\') {
-		*t++ = *p++;
-		if (p == pend) break;
+		if (++p == pend) break;
 	    }
-	    *t++ = *p++;
+	    p++;
 	}
-	*t = '\0';
 	if (maxnest == 0) {
 	    status = push_globs(ary, buf, flags);
 	    if (status) break;
@@ -1281,9 +1316,33 @@ dir_s_aref(obj, str)
  *  Returns the filenames found by expanding the pattern given in
  *  <i>string</i>, either as an <i>array</i> or as parameters to the
  *  block. Note that this pattern is not a regexp (it's closer to a
- *  shell glob). See <code>File::fnmatch</code> for
- *  details of file name matching and the meaning of the <i>flags</i>
- *  parameter.
+ *  shell glob). See <code>File::fnmatch</code> for the meaning of
+ *  the <i>flags</i> parameter.
+ *
+ *  <code>*</code>::        Matches any file. Can be restricted by
+ *                          other values in the glob. <code>*</code>
+ *                          will match all files; <code>c*</code> will
+ *                          match all files beginning with
+ *                          <code>c</code>; <code>*c</code> will match
+ *                          all files ending with <code>c</code>; and
+ *                          <code>*c*</code> will match all files that
+ *                          have <code>c</code> in them (including at
+ *                          the beginning or end). Equivalent to
+ *                          <code>/ .* /x</code> in regexp.
+ *  <code>**</code>::       Matches directories recursively.
+ *  <code>?</code>::        Matches any one character. Equivalent to
+ *                          <code>/.{1}/</code> in regexp.
+ *  <code>[set]</code>::    Matches any one character in +set+.
+ *                          Behaves exactly like character sets in
+ *                          Regexp, including set negation
+ *                          (<code>[^a-z]</code>).
+ *  <code>{p,q}</code>::    Matches either literal <code>p</code> or
+ *                          literal <code>q</code>. Matching literals
+ *                          may be more than one character in length.
+ *                          More than two literals may be specified.
+ *                          Equivalent to pattern alternation in
+ *                          regexp.
+ *  <code>\</code>::        Escapes the next metacharacter.
  *
  *     Dir["config.?"]                     #=> ["config.h"]
  *     Dir.glob("config.?")                #=> ["config.h"]
@@ -1293,6 +1352,19 @@ dir_s_aref(obj, str)
  *     Dir.glob("*")                       #=> ["config.h", "main.rb"]
  *     Dir.glob("*", File::FNM_DOTMATCH)   #=> [".", "..", "config.h", "main.rb"]
  *
+ *     rbfiles = File.join("**", "*.rb")
+ *     Dir.glob(rbfiles)                   #=> ["main.rb",
+ *                                              "lib/song.rb",
+ *                                              "lib/song/karaoke.rb"]
+ *     libdirs = File.join("**", "lib")
+ *     Dir.glob(libdirs)                   #=> ["lib"]
+ *
+ *     librbfiles = File.join("**", "lib", "**", "*.rb")
+ *     Dir.glob(librbfiles)                #=> ["lib/song.rb",
+ *                                              "lib/song/karaoke.rb"]
+ *
+ *     librbfiles = File.join("**", "lib", "*.rb")
+ *     Dir.glob(librbfiles)                #=> ["lib/song.rb"]
  */
 static VALUE
 dir_s_glob(argc, argv, obj)
@@ -1315,7 +1387,6 @@ static VALUE
 dir_open_dir(path)
     VALUE path;
 {
-    struct dir_data *dp;
     VALUE dir = rb_funcall(rb_cDir, rb_intern("open"), 1, path);
 
     if (TYPE(dir) != T_DATA ||
@@ -1386,8 +1457,27 @@ dir_entries(io, dirname)
  *  similar to shell filename globbing. It may contain the following
  *  metacharacters:
  *
- *  <i>flags</i> is a bitwise OR of the <code>FNM_xxx</code> parameters.
- *  The same glob pattern and flags are used by <code>Dir::glob</code>.
+ *  <code>*</code>::        Matches any file. Can be restricted by
+ *                          other values in the glob. <code>*</code>
+ *                          will match all files; <code>c*</code> will
+ *                          match all files beginning with
+ *                          <code>c</code>; <code>*c</code> will match
+ *                          all files ending with <code>c</code>; and
+ *                          <code>*c*</code> will match all files that
+ *                          have <code>c</code> in them (including at
+ *                          the beginning or end). Equivalent to
+ *                          <code>/ .* /x</code> in regexp.
+ *  <code>?</code>::        Matches any one character. Equivalent to
+ *                          <code>/.{1}/</code> in regexp.
+ *  <code>[set]</code>::    Matches any one character in +set+.
+ *                          Behaves exactly like character sets in
+ *                          Regexp, including set negation
+ *                          (<code>[^a-z]</code>).
+ *  <code>\</code>::        Escapes the next metacharacter.
+ *
+ *  <i>flags</i> is a bitwise OR of the <code>FNM_xxx</code>
+ *  parameters. The same glob pattern and flags are used by
+ *  <code>Dir::glob</code>.
  *
  *     File.fnmatch('cat',       'cat')        #=> true
  *     File.fnmatch('cat',       'category')   #=> false
