@@ -2,8 +2,8 @@
 
   file.c -
 
-  $Author: nobu $
-  $Date: 2005/10/18 23:29:24 $
+  $Author: eban $
+  $Date: 2006/02/06 07:13:53 $
   created at: Mon Nov 15 12:24:34 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -71,6 +71,33 @@ char *strrchr _((const char*,const char));
 # define fseeko  fseek
 #endif
 
+#ifdef __BEOS__ /* should not change ID if -1 */
+static int
+be_chown(const char *path, uid_t owner, gid_t group)
+{
+    if (owner == -1 || group == -1) {
+	struct stat st;
+	if (stat(path, &st) < 0) return -1;
+	if (owner == -1) owner = st.st_uid;
+	if (group == -1) group = st.st_gid;
+    }
+    return chown(path, owner, group);
+}
+#define chown be_chown
+static int
+be_fchown(int fd, uid_t owner, gid_t group)
+{
+    if (owner == -1 || group == -1) {
+	struct stat st;
+	if (fstat(fd, &st) < 0) return -1;
+	if (owner == -1) owner = st.st_uid;
+	if (group == -1) group = st.st_gid;
+    }
+    return fchown(fd, owner, group);
+}
+#define fchown be_fchown
+#endif /* __BEOS__ */
+
 VALUE rb_cFile;
 VALUE rb_mFileTest;
 static VALUE rb_cStat;
@@ -635,6 +662,40 @@ rb_stat(file, st)
     return stat(StringValueCStr(file), st);
 }
 
+#ifdef _WIN32
+static HANDLE
+w32_io_info(file, st)
+    VALUE *file;
+    BY_HANDLE_FILE_INFORMATION *st;
+{
+    VALUE tmp;
+    HANDLE f, ret = 0;
+
+    tmp = rb_check_convert_type(*file, T_FILE, "IO", "to_io");
+    if (!NIL_P(tmp)) {
+	OpenFile *fptr;
+
+	GetOpenFile(tmp, fptr);
+	f = (HANDLE)rb_w32_get_osfhandle(fileno(fptr->f));
+	if (f == (HANDLE)-1) return INVALID_HANDLE_VALUE;
+    }
+    else {
+	SafeStringValue(*file);
+	f = CreateFile(StringValueCStr(*file), 0,
+	               FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING,
+	               rb_w32_iswin95() ? 0 : FILE_FLAG_BACKUP_SEMANTICS, NULL);
+	if (f == INVALID_HANDLE_VALUE) return f;
+	ret = f;
+    }
+    if (GetFileType(f) == FILE_TYPE_DISK) {
+	ZeroMemory(st, sizeof(*st));
+	if (GetFileInformationByHandle(f, st)) return ret;
+    }
+    if (ret) CloseHandle(ret);
+    return INVALID_HANDLE_VALUE;
+}
+#endif
+
 /*
  *  call-seq:
  *     File.stat(file_name)   =>  stat
@@ -788,6 +849,7 @@ group_member(gid)
 #  define S_IXUGO		(S_IXUSR | S_IXGRP | S_IXOTH)
 #endif
 
+#ifndef HAVE_EACCESS
 int
 eaccess(path, mode)
      const char *path;
@@ -823,9 +885,13 @@ eaccess(path, mode)
 
     return -1;
 #else
+# if _MSC_VER >= 1400
+    mode &= 6;
+# endif
     return access(path, mode);
 #endif
 }
+#endif
 
 
 /*
@@ -1282,7 +1348,7 @@ check3rdbyte(fname, mode)
  * call-seq:
  *   File.setuid?(file_name)   =>  true or false
  *
- * Returns <code>true</code> if the named file is a has the setuid bit set.
+ * Returns <code>true</code> if the named file has the setuid bit set.
  */
 
 static VALUE
@@ -1300,7 +1366,7 @@ test_suid(obj, fname)
  * call-seq:
  *   File.setgid?(file_name)   =>  true or false
  *
- * Returns <code>true</code> if the named file is a has the setgid bit set.
+ * Returns <code>true</code> if the named file has the setgid bit set.
  */
 
 static VALUE
@@ -1318,7 +1384,7 @@ test_sgid(obj, fname)
  * call-seq:
  *   File.sticky?(file_name)   =>  true or false
  *
- * Returns <code>true</code> if the named file is a has the sticky bit set.
+ * Returns <code>true</code> if the named file has the sticky bit set.
  */
 
 static VALUE
@@ -1332,6 +1398,71 @@ test_sticky(obj, fname)
 #endif
 }
 
+/*
+ * call-seq:
+ *   File.identical?(file_1, file_2)   =>  true or false
+ *
+ * Returns <code>true</code> if the named files are identical.
+ *
+ *     open("a", "w") {}
+ *     p File.identical?("a", "a")      #=> true
+ *     p File.identical?("a", "./a")    #=> true
+ *     File.link("a", "b")
+ *     p File.identical?("a", "b")      #=> true
+ *     File.symlink("a", "c")
+ *     p File.identical?("a", "c")      #=> true
+ *     open("d", "w") {}
+ *     p File.identical?("a", "d")      #=> false
+ */
+
+static VALUE
+test_identical(obj, fname1, fname2)
+    VALUE obj, fname1, fname2;
+{
+#ifndef DOSISH
+    struct stat st1, st2;
+
+    if (rb_stat(fname1, &st1) < 0) return Qfalse;
+    if (rb_stat(fname2, &st2) < 0) return Qfalse;
+    if (st1.st_dev != st2.st_dev) return Qfalse;
+    if (st1.st_ino != st2.st_ino) return Qfalse;
+#else
+#ifdef _WIN32
+    BY_HANDLE_FILE_INFORMATION st1, st2;
+    HANDLE f1 = 0, f2 = 0;
+#endif
+
+    rb_secure(2);
+#ifdef _WIN32
+    f1 = w32_io_info(&fname1, &st1);
+    if (f1 == INVALID_HANDLE_VALUE) return Qfalse;
+    f2 = w32_io_info(&fname2, &st2);
+    if (f1) CloseHandle(f1);
+    if (f2 == INVALID_HANDLE_VALUE) return Qfalse;
+    if (f2) CloseHandle(f2);
+
+    if (st1.dwVolumeSerialNumber == st2.dwVolumeSerialNumber &&
+	st1.nFileIndexHigh == st2.nFileIndexHigh &&
+	st1.nFileIndexLow == st2.nFileIndexLow)
+	return Qtrue;
+    if (!f1 || !f2) return Qfalse;
+    if (rb_w32_iswin95()) return Qfalse;
+#else
+    SafeStringValue(fname1);
+    fname1 = rb_str_new4(fname1);
+    SafeStringValue(fname2);
+    if (access(RSTRING(fname1)->ptr, 0)) return Qfalse;
+    if (access(RSTRING(fname2)->ptr, 0)) return Qfalse;
+#endif
+    fname1 = rb_file_expand_path(fname1, Qnil);
+    fname2 = rb_file_expand_path(fname2, Qnil);
+    if (RSTRING(fname1)->len != RSTRING(fname2)->len) return Qfalse;
+    if (rb_memcicmp(RSTRING(fname1)->ptr, RSTRING(fname2)->ptr, RSTRING(fname1)->len))
+	return Qfalse;
+#endif
+    return Qtrue;
+}
+
 /*
  * call-seq:
  *    File.size(file_name)   => integer
@@ -1566,7 +1697,7 @@ chmod_internal(path, mode)
     const char *path;
     void *mode;
 {
-    if (chmod(path, (int)mode) < 0)
+    if (chmod(path, *(int *)mode) < 0)
 	rb_sys_fail(path);
 }
 
@@ -1597,7 +1728,7 @@ rb_file_s_chmod(argc, argv)
     rb_scan_args(argc, argv, "1*", &vmode, &rest);
     mode = NUM2INT(vmode);
 
-    n = apply2files(chmod_internal, rest, (void *)(long)mode);
+    n = apply2files(chmod_internal, rest, &mode);
     return LONG2FIX(n);
 }
 
@@ -1957,6 +2088,7 @@ static VALUE
 rb_file_s_link(klass, from, to)
     VALUE klass, from, to;
 {
+#ifdef HAVE_LINK
     SafeStringValue(from);
     SafeStringValue(to);
 
@@ -1964,6 +2096,10 @@ rb_file_s_link(klass, from, to)
 	sys_fail2(from, to);
     }
     return INT2FIX(0);
+#else
+    rb_notimplement();
+    return Qnil;		/* not reached */
+#endif
 }
 
 /*
@@ -2019,7 +2155,11 @@ rb_file_s_readlink(klass, path)
 
     SafeStringValue(path);
     buf = xmalloc(size);
-    while ((rv = readlink(StringValueCStr(path), buf, size)) == size) {
+    while ((rv = readlink(RSTRING(path)->ptr, buf, size)) == size
+#ifdef _AIX
+	    || (rv < 0 && errno == ERANGE) /* quirky behavior of GPFS */
+#endif
+	) {
 	size *= 2;
 	buf = xrealloc(buf, size);
     }
@@ -2090,7 +2230,7 @@ rb_file_s_rename(klass, from, to)
     dst = StringValueCStr(to);
     if (rename(src, dst) < 0) {
 #if defined __CYGWIN__
-	extern unsigned long __attribute__((stdcall)) GetLastError();
+	extern unsigned long __attribute__((stdcall)) GetLastError(void);
 	errno = GetLastError(); /* This is a Cygwin bug */
 #elif defined DOSISH && !defined _WIN32
 	if (errno == EEXIST
@@ -2230,7 +2370,11 @@ rb_path_next(s)
     return (char *)s;
 }
 
+#if defined(DOSISH_UNC) || defined(DOSISH_DRIVE_LETTER) 
 #define skipprefix rb_path_skip_prefix
+#else
+#define skipprefix(path) (path)
+#endif
 char *
 rb_path_skip_prefix(path)
     const char *path;
@@ -2238,7 +2382,9 @@ rb_path_skip_prefix(path)
 #if defined(DOSISH_UNC) || defined(DOSISH_DRIVE_LETTER) 
 #ifdef DOSISH_UNC
     if (isdirsep(path[0]) && isdirsep(path[1])) {
-	if (*(path = nextdirsep(path + 2)))
+	path += 2;
+	while (isdirsep(*path)) path++;
+	if (*(path = nextdirsep(path)) && path[1] && !isdirsep(path[1]))
 	    path = nextdirsep(path + 1);
 	return (char *)path;
     }
@@ -2594,6 +2740,9 @@ rb_file_s_basename(argc, argv)
 {
     VALUE fname, fext, basename;
     char *name, *p;
+#if defined DOSISH_DRIVE_LETTER || defined DOSISH_UNC
+    char *root;
+#endif
     int f;
 
     if (rb_scan_args(argc, argv, "11", &fname, &fext) == 2) {
@@ -2602,14 +2751,30 @@ rb_file_s_basename(argc, argv)
     StringValue(fname);
     if (RSTRING(fname)->len == 0 || !*(name = RSTRING(fname)->ptr))
 	return fname;
-    if (!*(name = skiproot(name))) {
+    name = skipprefix(name);
+#if defined DOSISH_DRIVE_LETTER || defined DOSISH_UNC
+    root = name;
+#endif
+    while (isdirsep(*name))
+	name++;
+    if (!*name) {
 	p = name - 1;
 	f = 1;
+#if defined DOSISH_DRIVE_LETTER || defined DOSISH_UNC
+	if (name != root) {
+	    /* has slashes */
+	}
 #ifdef DOSISH_DRIVE_LETTER
-	if (*p == ':') {
+	else if (*p == ':') {
 	    p++;
 	    f = 0;
 	}
+#endif
+#ifdef DOSISH_UNC
+	else {
+	    p = "/";
+	}
+#endif
 #endif
     }
     else if (!(p = strrdirsep(name))) {
@@ -2664,9 +2829,16 @@ rb_file_s_dirname(klass, fname)
     }
     if (p == name)
 	return rb_str_new2(".");
+#ifdef DOSISH_DRIVE_LETTER
+    if (has_drive_letter(name) && isdirsep(*(name + 2))) {
+	dirname = rb_str_new(name, 3);
+	rb_str_cat(dirname, skiproot(name + 2), p - skiproot(name + 2));
+    }
+    else
+#endif
     dirname = rb_str_new(name, p - name);
 #ifdef DOSISH_DRIVE_LETTER
-    if (root == name + 2 && name[1] == ':')
+    if (has_drive_letter(name) && root == name + 2 && p - name == 2)
 	rb_str_cat(dirname, ".", 1);
 #endif
     OBJ_INFECT(dirname, fname);
@@ -2935,7 +3107,22 @@ rb_file_truncate(obj, len)
 #  define LOCK_UN 8
 # endif
 
-#if 1
+#ifdef __CYGWIN__
+static int
+#include <winerror.h>
+cygwin_flock(int fd, int op)
+{
+    int old_errno = errno;
+    int ret = flock(fd, op);
+    if (GetLastError() == ERROR_NOT_LOCKED) {
+	ret = 0;
+	errno = old_errno;
+    }
+    return ret;
+}
+# define flock(fd, op) cygwin_flock(fd, op)
+#endif
+
 static int
 rb_thread_flock(fd, op, fptr)
     int fd, op;
@@ -2965,8 +3152,10 @@ rb_thread_flock(fd, op, fptr)
     }
     return 0;
 }
-#define flock(fd, op) rb_thread_flock(fd, op, fptr)
+#ifdef __CYGWIN__
+# undef flock
 #endif
+#define flock(fd, op) rb_thread_flock(fd, op, fptr)
 
 /*
  *  call-seq:
@@ -3108,7 +3297,7 @@ test_check(n, argc, argv)
  *
  * Tests that take two files:
  *
- *     ?-  | boolean | True if file1 is a hard link to file2
+ *     ?-  | boolean | True if file1 and file2 are identical
  *     ?=  | boolean | True if the modification times of file1
  *         |         | and file2 are equal
  *     ?<  | boolean | True if the modification time of file1
@@ -3222,7 +3411,12 @@ rb_f_test(argc, argv)
 	}
     }
 
-    if (strchr("-=<>", cmd)) {
+    if (cmd == '-') {
+	CHECK(2);
+	return test_identical(0, argv[1], argv[2]);
+    }
+
+    if (strchr("=<>", cmd)) {
 	struct stat st1, st2;
 
 	CHECK(2);
@@ -3230,11 +3424,6 @@ rb_f_test(argc, argv)
 	if (rb_stat(argv[2], &st2) < 0) return Qfalse;
 
 	switch (cmd) {
-	  case '-':
-	    if (st1.st_dev == st2.st_dev && st1.st_ino == st2.st_ino)
-		return Qtrue;
-	    return Qfalse;
-
 	  case '=':
 	    if (st1.st_mtime == st2.st_mtime) return Qtrue;
 	    return Qfalse;
@@ -4194,6 +4383,8 @@ Init_File()
     define_filetest_function("setgid?", test_sgid, 1);
     define_filetest_function("sticky?", test_sticky, 1);
 
+    define_filetest_function("identical?", test_identical, 2);
+
     rb_define_singleton_method(rb_cFile, "stat",  rb_file_s_stat, 1);
     rb_define_singleton_method(rb_cFile, "lstat", rb_file_s_lstat, 1);
     rb_define_singleton_method(rb_cFile, "ftype", rb_file_s_ftype, 1);
