@@ -2,8 +2,8 @@
 
   gc.c -
 
-  $Author: matz $
-  $Date: 2005/01/20 09:34:36 $
+  $Author: akr $
+  $Date: 2006/03/04 06:28:51 $
   created at: Tue Oct  5 09:44:46 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -185,14 +185,14 @@ static st_table *finalizer_table = 0;
 /*
  *  call-seq:
  *     GC.enable    => true or false
- *  
+ *
  *  Enables garbage collection, returning <code>true</code> if garbage
  *  collection was previously disabled.
- *     
+ *
  *     GC.disable   #=> false
  *     GC.enable    #=> true
  *     GC.enable    #=> false
- *     
+ *
  */
 
 VALUE
@@ -207,13 +207,13 @@ rb_gc_enable()
 /*
  *  call-seq:
  *     GC.disable    => true or false
- *  
+ *
  *  Disables garbage collection, returning <code>true</code> if garbage
  *  collection was already disabled.
- *     
+ *
  *     GC.disable   #=> false
  *     GC.disable   #=> true
- *     
+ *
  */
 
 VALUE
@@ -276,6 +276,10 @@ rb_global_variable(var)
     rb_gc_register_address(var);
 }
 
+#if defined(_MSC_VER) || defined(__BORLANDC__) || defined(__CYGWIN__)
+#pragma pack(push, 1) /* magic for reducing sizeof(RVALUE): 24 -> 20 */
+#endif
+
 typedef struct RVALUE {
     union {
 	struct {
@@ -296,7 +300,7 @@ typedef struct RVALUE {
 	struct RFile   file;
 	struct RNode   node;
 	struct RMatch  match;
-	struct RVarmap varmap; 
+	struct RVarmap varmap;
 	struct SCOPE   scope;
     } as;
 #ifdef GC_DEBUG
@@ -305,11 +309,16 @@ typedef struct RVALUE {
 #endif
 } RVALUE;
 
+#if defined(_MSC_VER) || defined(__BORLANDC__) || defined(__CYGWIN__)
+#pragma pack(pop)
+#endif
+
 static RVALUE *freelist = 0;
 static RVALUE *deferred_final_list = 0;
 
 #define HEAPS_INCREMENT 10
 static struct heaps_slot {
+    void *membase;
     RVALUE *slot;
     int limit;
 } *heaps;
@@ -347,8 +356,7 @@ add_heap()
     }
 
     for (;;) {
-	RUBY_CRITICAL(p = heaps[heaps_used].slot = (RVALUE*)malloc(sizeof(RVALUE)*heap_slots));
-	heaps[heaps_used].limit = heap_slots;
+	RUBY_CRITICAL(p = (RVALUE*)malloc(sizeof(RVALUE)*(heap_slots+1)));
 	if (p == 0) {
 	    if (heap_slots == HEAP_MIN_SLOTS) {
 		rb_memerror();
@@ -356,6 +364,13 @@ add_heap()
 	    heap_slots = HEAP_MIN_SLOTS;
 	    continue;
 	}
+        heaps[heaps_used].membase = p;
+        if ((VALUE)p % sizeof(RVALUE) == 0)
+            heap_slots += 1;
+        else
+            p = (RVALUE*)((VALUE)p + sizeof(RVALUE) - ((VALUE)p % sizeof(RVALUE)));
+        heaps[heaps_used].slot = p;
+        heaps[heaps_used].limit = heap_slots;
 	break;
     }
     pend = p + heap_slots;
@@ -363,6 +378,7 @@ add_heap()
     if (himem < pend) himem = pend;
     heaps_used++;
     heap_slots *= 1.8;
+    if (heap_slots <= 0) heap_slots = HEAP_MIN_SLOTS;
 
     while (p < pend) {
 	p->as.free.flags = 0;
@@ -603,12 +619,12 @@ is_pointer_to_heap(ptr)
     register long i;
 
     if (p < lomem || p > himem) return Qfalse;
+    if ((VALUE)p % sizeof(RVALUE) != 0) return Qfalse;
 
     /* check if p looks like a pointer */
     for (i=0; i < heaps_used; i++) {
 	heap_org = heaps[i].slot;
-	if (heap_org <= p && p < heap_org + heaps[i].limit &&
-	    ((((char*)p)-((char*)heap_org))%sizeof(RVALUE)) == 0)
+	if (heap_org <= p && p < heap_org + heaps[i].limit)
 	    return Qtrue;
     }
     return Qfalse;
@@ -649,7 +665,7 @@ mark_entry(key, value, lev)
     return ST_CONTINUE;
 }
 
-void
+static void
 mark_tbl(tbl, lev)
     st_table *tbl;
     int lev;
@@ -676,7 +692,7 @@ mark_keyvalue(key, value, lev)
     return ST_CONTINUE;
 }
 
-void
+static void
 mark_hash(tbl, lev)
     st_table *tbl;
     int lev;
@@ -703,7 +719,7 @@ rb_gc_mark_maybe(obj)
 
 #define GC_LEVEL_MAX 250
 
-void
+static void
 gc_mark(ptr, lev)
     VALUE ptr;
     int lev;
@@ -713,7 +729,7 @@ gc_mark(ptr, lev)
     obj = RANY(ptr);
     if (rb_special_const_p(ptr)) return; /* special const not marked */
     if (obj->as.basic.flags == 0) return;       /* free cell */
-    if (obj->as.basic.flags & FL_MARK) return;  /* already marked */ 
+    if (obj->as.basic.flags & FL_MARK) return;  /* already marked */
     obj->as.basic.flags |= FL_MARK;
 
     if (lev > GC_LEVEL_MAX || (lev == 0 && ruby_stack_check())) {
@@ -751,7 +767,7 @@ gc_mark_children(ptr, lev)
     obj = RANY(ptr);
     if (rb_special_const_p(ptr)) return; /* special const not marked */
     if (obj->as.basic.flags == 0) return;       /* free cell */
-    if (obj->as.basic.flags & FL_MARK) return;  /* already marked */ 
+    if (obj->as.basic.flags & FL_MARK) return;  /* already marked */
     obj->as.basic.flags |= FL_MARK;
 
   marking:
@@ -814,6 +830,9 @@ gc_mark_children(ptr, lev)
 	  case NODE_OP_ASGN_OR:
 	  case NODE_OP_ASGN_AND:
 	  case NODE_MODULE:
+	  case NODE_ALIAS:
+	  case NODE_VALIAS:
+	  case NODE_ARGS:
 	    gc_mark((VALUE)obj->as.node.u1.node, lev);
 	    /* fall through */
 	  case NODE_METHOD:	/* 2 */
@@ -828,6 +847,7 @@ gc_mark_children(ptr, lev)
 	  case NODE_COLON3:
 	  case NODE_OPT_N:
 	  case NODE_EVSTR:
+	  case NODE_UNDEF:
 	    ptr = (VALUE)obj->as.node.u2.node;
 	    goto again;
 
@@ -842,7 +862,6 @@ gc_mark_children(ptr, lev)
 	  case NODE_NEXT:
 	  case NODE_YIELD:
 	  case NODE_COLON2:
-	  case NODE_ARGS:
 	  case NODE_SPLAT:
 	  case NODE_TO_ARY:
 	  case NODE_SVALUE:
@@ -867,11 +886,8 @@ gc_mark_children(ptr, lev)
 	  case NODE_CVAR:
 	  case NODE_NTH_REF:
 	  case NODE_BACK_REF:
-	  case NODE_ALIAS:
-	  case NODE_VALIAS:
 	  case NODE_REDO:
 	  case NODE_RETRY:
-	  case NODE_UNDEF:
 	  case NODE_SELF:
 	  case NODE_NIL:
 	  case NODE_TRUE:
@@ -880,13 +896,11 @@ gc_mark_children(ptr, lev)
 	  case NODE_BLOCK_ARG:
 	  case NODE_POSTEXE:
 	    break;
-#ifdef C_ALLOCA
 	  case NODE_ALLOCA:
 	    mark_locations_array((VALUE*)obj->as.node.u1.value,
 				 obj->as.node.u3.cnt);
 	    ptr = (VALUE)obj->as.node.u2.node;
 	    goto again;
-#endif
 
 	  default:		/* unlisted NODE */
 	    if (is_pointer_to_heap(obj->as.node.u1.node)) {
@@ -1021,7 +1035,7 @@ free_unused_heaps()
 
     for (i = j = 1; j < heaps_used; i++) {
 	if (heaps[i].limit == 0) {
-	    free(heaps[i].slot);
+	    free(heaps[i].membase);
 	    heaps_used--;
 	}
 	else {
@@ -1033,6 +1047,8 @@ free_unused_heaps()
     }
 }
 
+void rb_gc_abort_threads(void);
+
 static void
 gc_sweep()
 {
@@ -1055,7 +1071,9 @@ gc_sweep()
     }
 
     mark_source_filename(ruby_sourcefile);
-    st_foreach(source_filenames, sweep_source_filename, 0);
+    if (source_filenames) {
+        st_foreach(source_filenames, sweep_source_filename, 0);
+    }
 
     freelist = 0;
     final_list = deferred_final_list;
@@ -1085,7 +1103,7 @@ gc_sweep()
 	    }
 	    else if (RBASIC(p)->flags == FL_MARK) {
 		/* objects to be finalized */
-		/* do notning remain marked */
+		/* do nothing remain marked */
 	    }
 	    else {
 		RBASIC(p)->flags &= ~FL_MARK;
@@ -1230,11 +1248,9 @@ obj_free(obj)
 		RUBY_CRITICAL(free(RANY(obj)->as.node.u1.tbl));
 	    }
 	    break;
-#ifdef C_ALLOCA
 	  case NODE_ALLOCA:
 	    RUBY_CRITICAL(free(RANY(obj)->as.node.u1.node));
 	    break;
-#endif
 	}
 	return;			/* no need to free iv_tbl */
 
@@ -1256,8 +1272,8 @@ obj_free(obj)
 	break;
 
       default:
-	rb_bug("gc_sweep(): unknown data type 0x%lx(%ld)", obj,
-	       RANY(obj)->as.basic.flags & T_MASK);
+	rb_bug("gc_sweep(): unknown data type 0x%lx(0x%lx)",
+	       RANY(obj)->as.basic.flags & T_MASK, obj);
     }
 }
 
@@ -1265,7 +1281,6 @@ void
 rb_gc_mark_frame(frame)
     struct FRAME *frame;
 {
-    mark_locations_array(frame->argv, frame->argc);
     gc_mark((VALUE)frame->node, 0);
 }
 
@@ -1330,10 +1345,12 @@ garbage_collect()
     during_gc++;
 
     init_mark_stack();
-    
+
+    gc_mark((VALUE)ruby_current_node, 0);
+
     /* mark frame stack */
     for (frame = ruby_frame; frame; frame = frame->prev) {
-	rb_gc_mark_frame(frame); 
+	rb_gc_mark_frame(frame);
 	if (frame->tmp) {
 	    struct FRAME *tmp = frame->tmp;
 	    while (tmp) {
@@ -1400,16 +1417,20 @@ garbage_collect()
     rb_mark_generic_ivar_tbl();
 
     rb_gc_mark_parser();
-    
+
     /* gc_mark objects whose marking are not completed*/
-    while (!MARK_STACK_EMPTY){
-	if (mark_stack_overflow){
-	    gc_mark_all();
+    do {
+	while (!MARK_STACK_EMPTY) {
+	    if (mark_stack_overflow){
+		gc_mark_all();
+	    }
+	    else {
+		gc_mark_rest();
+	    }
 	}
-	else {
-	    gc_mark_rest();
-	}
-    }
+	rb_gc_abort_threads();
+    } while (!MARK_STACK_EMPTY);
+
     gc_sweep();
 }
 
@@ -1427,7 +1448,7 @@ rb_gc()
  *     ObjectSpace.garbage_collect  => nil
  *
  *  Initiates garbage collection, unless manually disabled.
- *     
+ *
  */
 
 VALUE
@@ -1483,14 +1504,6 @@ Init_stack(addr)
 	    STACK_LEVEL_MAX = (rlim.rlim_cur - space) / sizeof(VALUE);
 	}
     }
-#if defined(__ia64__) && (!defined(__GNUC__) || __GNUC__ < 2 || defined(__OPTIMIZE__))
-    /* ruby crashes on IA64 if compiled with optimizer on */
-    /* when if STACK_LEVEL_MAX is greater than this magic number */
-    /* I know this is a kludge.  I suspect optimizer bug */
-#define IA64_MAGIC_STACK_LIMIT 49152
-    if (STACK_LEVEL_MAX > IA64_MAGIC_STACK_LIMIT)
-	STACK_LEVEL_MAX = IA64_MAGIC_STACK_LIMIT;
-#endif
 #endif
 }
 
@@ -1501,29 +1514,29 @@ Init_stack(addr)
  *  The <code>ObjectSpace</code> module contains a number of routines
  *  that interact with the garbage collection facility and allow you to
  *  traverse all living objects with an iterator.
- *     
+ *
  *  <code>ObjectSpace</code> also provides support for object
  *  finalizers, procs that will be called when a specific object is
  *  about to be destroyed by garbage collection.
- *     
+ *
  *     include ObjectSpace
- *     
- *     
+ *
+ *
  *     a = "A"
  *     b = "B"
  *     c = "C"
- *     
- *     
+ *
+ *
  *     define_finalizer(a, proc {|id| puts "Finalizer one on #{id}" })
  *     define_finalizer(a, proc {|id| puts "Finalizer two on #{id}" })
  *     define_finalizer(b, proc {|id| puts "Finalizer three on #{id}" })
- *     
+ *
  *  <em>produces:</em>
- *     
+ *
  *     Finalizer three on 537763470
  *     Finalizer one on 537763480
  *     Finalizer two on 537763480
- *     
+ *
  */
 
 void
@@ -1605,7 +1618,7 @@ os_obj_of(of)
 /*
  *  call-seq:
  *     ObjectSpace.each_object([module]) {|obj| ... } => fixnum
- *  
+ *
  *  Calls the block once for each living, nonimmediate object in this
  *  Ruby process. If <i>module</i> is specified, calls the block
  *  for only those classes or modules that match (or are a subclass of)
@@ -1615,15 +1628,15 @@ os_obj_of(of)
  *  never returned. In the example below, <code>each_object</code>
  *  returns both the numbers we defined and several constants defined in
  *  the <code>Math</code> module.
- *     
+ *
  *     a = 102.7
  *     b = 95       # Won't be returned
  *     c = 12345678987654321
  *     count = ObjectSpace.each_object(Numeric) {|x| p x }
  *     puts "Total count: #{count}"
- *     
+ *
  *  <em>produces:</em>
- *     
+ *
  *     12345678987654321
  *     102.7
  *     2.71828182845905
@@ -1632,7 +1645,7 @@ os_obj_of(of)
  *     1.7976931348623157e+308
  *     2.2250738585072e-308
  *     Total count: 7
- *     
+ *
  */
 
 static VALUE
@@ -1708,9 +1721,9 @@ call_final(os, obj)
 /*
  *  call-seq:
  *     ObjectSpace.undefine_finalizer(obj)
- *  
+ *
  *  Removes all finalizers for <i>obj</i>.
- *     
+ *
  */
 
 static VALUE
@@ -1726,10 +1739,10 @@ undefine_final(os, obj)
 /*
  *  call-seq:
  *     ObjectSpace.define_finalizer(obj, aProc=proc())
- *  
- *  Adds <i>aProc</i> as a finalizer, to be called when <i>obj</i>
- *  is about to be destroyed.
- *     
+ *
+ *  Adds <i>aProc</i> as a finalizer, to be called after <i>obj</i>
+ *  was destroyed.
+ *
  */
 
 static VALUE
@@ -1793,19 +1806,22 @@ run_final(obj)
 {
     long i;
     int status, critical_save = rb_thread_critical;
-    VALUE args[3], table;
+    VALUE args[3], table, objid;
 
+    objid = rb_obj_id(obj);	/* make obj into id */
     rb_thread_critical = Qtrue;
-    args[1] = rb_ary_new3(1, rb_obj_id(obj)); /* make obj into id */
+    args[1] = 0;
     args[2] = (VALUE)ruby_safe_level;
     for (i=0; i<RARRAY(finalizers)->len; i++) {
 	args[0] = RARRAY(finalizers)->ptr[i];
+	if (!args[1]) args[1] = rb_ary_new3(1, objid);
 	rb_protect((VALUE(*)_((VALUE)))run_single_final, (VALUE)args, &status);
     }
     if (finalizer_table && st_delete(finalizer_table, (st_data_t*)&obj, &table)) {
 	for (i=0; i<RARRAY(table)->len; i++) {
 	    VALUE final = RARRAY(table)->ptr[i];
 	    args[0] = RARRAY(final)->ptr[1];
+	    if (!args[1]) args[1] = rb_ary_new3(1, objid);
 	    args[2] = FIX2INT(RARRAY(final)->ptr[0]);
 	    rb_protect((VALUE(*)_((VALUE)))run_single_final, (VALUE)args, &status);
 	}
@@ -1833,7 +1849,9 @@ rb_gc_call_finalizer_at_exit()
 
     /* run finalizers */
     if (need_call_final) {
-	finalize_list(deferred_final_list);
+	p = deferred_final_list;
+	deferred_final_list = 0;
+	finalize_list(p);
 	for (i = 0; i < heaps_used; i++) {
 	    p = heaps[i].slot; pend = p + heaps[i].limit;
 	    while (p < pend) {
@@ -1872,33 +1890,37 @@ rb_gc_call_finalizer_at_exit()
 /*
  *  call-seq:
  *     ObjectSpace._id2ref(object_id) -> an_object
- *  
+ *
  *  Converts an object id to a reference to the object. May not be
  *  called on an object id passed as a parameter to a finalizer.
- *     
+ *
  *     s = "I am a string"                    #=> "I am a string"
  *     r = ObjectSpace._id2ref(s.object_id)   #=> "I am a string"
  *     r == s                                 #=> true
- *     
+ *
  */
 
 static VALUE
-id2ref(obj, id)
-    VALUE obj, id;
+id2ref(obj, objid)
+    VALUE obj, objid;
 {
     unsigned long ptr, p0;
 
     rb_secure(4);
-    p0 = ptr = NUM2ULONG(id);
+    p0 = ptr = NUM2ULONG(objid);
     if (ptr == Qtrue) return Qtrue;
     if (ptr == Qfalse) return Qfalse;
     if (ptr == Qnil) return Qnil;
     if (FIXNUM_P(ptr)) return (VALUE)ptr;
-    if (SYMBOL_P(ptr) && rb_id2name(SYM2ID((VALUE)ptr)) != 0) {
-	return (VALUE)ptr;
+    ptr = objid ^ FIXNUM_FLAG;	/* unset FIXNUM_FLAG */
+
+    if ((ptr % sizeof(RVALUE)) == (4 << 2)) {
+        ID symid = ptr / sizeof(RVALUE);
+        if (rb_id2name(symid) == 0)
+            rb_raise(rb_eRangeError, "%p is not symbol id value", p0);
+        return ID2SYM(symid);
     }
 
-    ptr = id ^ FIXNUM_FLAG;	/* unset FIXNUM_FLAG */
     if (!is_pointer_to_heap((void *)ptr)|| BUILTIN_TYPE(ptr) >= T_BLKTAG) {
 	rb_raise(rb_eRangeError, "0x%lx is not id value", p0);
     }
@@ -1908,6 +1930,73 @@ id2ref(obj, id)
     return (VALUE)ptr;
 }
 
+/*
+ *  Document-method: __id__
+ *  Document-method: object_id
+ *
+ *  call-seq:
+ *     obj.__id__       => fixnum
+ *     obj.object_id    => fixnum
+ *  
+ *  Returns an integer identifier for <i>obj</i>. The same number will
+ *  be returned on all calls to <code>id</code> for a given object, and
+ *  no two active objects will share an id.
+ *  <code>Object#object_id</code> is a different concept from the
+ *  <code>:name</code> notation, which returns the symbol id of
+ *  <code>name</code>. Replaces the deprecated <code>Object#id</code>.
+ */
+
+/*
+ *  call-seq:
+ *     obj.hash    => fixnum
+ *  
+ *  Generates a <code>Fixnum</code> hash value for this object. This
+ *  function must have the property that <code>a.eql?(b)</code> implies
+ *  <code>a.hash == b.hash</code>. The hash value is used by class
+ *  <code>Hash</code>. Any hash value that exceeds the capacity of a
+ *  <code>Fixnum</code> will be truncated before being used.
+ */
+
+VALUE
+rb_obj_id(VALUE obj)
+{
+    /*
+     *                32-bit VALUE space
+     *          MSB ------------------------ LSB
+     *  false   00000000000000000000000000000000
+     *  true    00000000000000000000000000000010
+     *  nil     00000000000000000000000000000100
+     *  undef   00000000000000000000000000000110
+     *  symbol  ssssssssssssssssssssssss00001110
+     *  object  oooooooooooooooooooooooooooooo00        = 0 (mod sizeof(RVALUE))
+     *  fixnum  fffffffffffffffffffffffffffffff1
+     *
+     *                    object_id space
+     *                                       LSB
+     *  false   00000000000000000000000000000000
+     *  true    00000000000000000000000000000010
+     *  nil     00000000000000000000000000000100
+     *  undef   00000000000000000000000000000110
+     *  symbol   000SSSSSSSSSSSSSSSSSSSSSSSSSSS0        S...S % A = 4 (S...S = s...s * A + 4)
+     *  object   oooooooooooooooooooooooooooooo0        o...o % A = 0
+     *  fixnum  fffffffffffffffffffffffffffffff1        bignum if required
+     *
+     *  where A = sizeof(RVALUE)/4
+     *
+     *  sizeof(RVALUE) is
+     *  20 if 32-bit, double is 4-byte aligned
+     *  24 if 32-bit, double is 8-byte aligned
+     *  40 if 64-bit
+     */
+    if (TYPE(obj) == T_SYMBOL) {
+        return (SYM2ID(obj) * sizeof(RVALUE) + (4 << 2)) | FIXNUM_FLAG;
+    }
+    if (SPECIAL_CONST_P(obj)) {
+        return LONG2NUM((long)obj);
+    }
+    return (VALUE)((long)obj|FIXNUM_FLAG);
+}
+
 /*
  *  The <code>GC</code> module provides an interface to Ruby's mark and
  *  sweep garbage collection mechanism. Some of the underlying methods
@@ -1945,6 +2034,10 @@ Init_GC()
 
     source_filenames = st_init_strtable();
 
-    nomem_error = rb_exc_new2(rb_eNoMemError, "failed to allocate memory");
     rb_global_variable(&nomem_error);
+    nomem_error = rb_exc_new2(rb_eNoMemError, "failed to allocate memory");
+
+    rb_define_method(rb_mKernel, "hash", rb_obj_id, 0);
+    rb_define_method(rb_mKernel, "__id__", rb_obj_id, 0);
+    rb_define_method(rb_mKernel, "object_id", rb_obj_id, 0);
 }
