@@ -2,8 +2,8 @@
 
   hash.c -
 
-  $Author: matz $
-  $Date: 2005/02/28 02:45:20 $
+  $Author: ocean $
+  $Date: 2006/02/01 13:27:47 $
   created at: Mon Nov 22 18:51:18 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -880,8 +880,6 @@ static VALUE
 rb_hash_clear(hash)
     VALUE hash;
 {
-    void *tmp;
-
     rb_hash_modify(hash);
     if (RHASH(hash)->tbl->num_entries > 0) {
 	rb_hash_foreach(hash, clear_i, 0);
@@ -1814,7 +1812,7 @@ ruby_setenv(name, value)
      *         RTL's environ global variable directly yet.
      */
     SetEnvironmentVariable(name,value);
-#elif defined __CYGWIN__
+#elif defined(HAVE_SETENV) && defined(HAVE_UNSETENV)
 #undef setenv
 #undef unsetenv
     if (value)
@@ -1822,7 +1820,7 @@ ruby_setenv(name, value)
     else
 	unsetenv(name);
 #else  /* WIN32 */
-
+    size_t len;
     int i=envix(name);		        /* where does it go? */
 
     if (environ == origenviron) {	/* need we copy environment? */
@@ -1837,30 +1835,28 @@ ruby_setenv(name, value)
 	tmpenv[max] = 0;
 	environ = tmpenv;		/* tell exec where it is now */
     }
-    if (!value) {
-	if (environ != origenviron) {
-	    char **envp = origenviron;
-	    while (*envp && *envp != environ[i]) envp++;
-	    if (!*envp)
-		free(environ[i]);
+    if (environ[i]) {
+	char **envp = origenviron;
+	while (*envp && *envp != environ[i]) envp++;
+	if (!*envp)
+	    free(environ[i]);
+	if (!value) {
+	    while (environ[i]) {
+		environ[i] = environ[i+1];
+		i++;
+	    }
+	    return;
 	}
-	while (environ[i]) {
-	    environ[i] = environ[i+1];
-	    i++;
-	}
-	return;
     }
-    if (!environ[i]) {			/* does not exist yet */
+    else {			/* does not exist yet */
+	if (!value) return;
 	REALLOC_N(environ, char*, i+2);	/* just expand it a bit */
 	environ[i+1] = 0;	/* make sure it's null terminated */
     }
-    else {
-	if (environ[i] != origenviron[i])
-	    free(environ[i]);
-    }
-    environ[i] = ALLOC_N(char, strlen(name) + strlen(value) + 2);
+    len = strlen(name) + strlen(value) + 2;
+    environ[i] = ALLOC_N(char, len);
 #ifndef MSDOS
-    sprintf(environ[i],"%s=%s",name,value); /* all that work just for this */
+    snprintf(environ[i],len,"%s=%s",name,value); /* all that work just for this */
 #else
     /* MS-DOS requires environment variable names to be in uppercase */
     /* [Tom Dinger, 27 August 1990: Well, it doesn't _require_ it, but
