@@ -284,7 +284,7 @@ class CGI
   # Standard internet newline sequence
   EOL = CR + LF
 
-  REVISION = '$Id: cgi.rb,v 1.68.2.5 2004/07/14 14:22:11 matz Exp $' #:nodoc:
+  REVISION = '$Id: cgi.rb,v 1.68.2.13 2006/05/18 08:14:37 matz Exp $' #:nodoc:
 
   NEEDS_BINMODE = true if /WIN/ni.match(RUBY_PLATFORM) 
 
@@ -716,6 +716,8 @@ class CGI
       when /shift_jis/ni
         content = NKF::nkf('-s', content)
         options["language"] = "ja" unless options.has_key?("language")
+      when /utf-8/ni
+        content = NKF::nkf('-w', content)
       end
     end
 
@@ -770,7 +772,7 @@ class CGI
   #   cookie1.domain  = 'domain'
   #   cookie1.expires = Time.now + 30
   #   cookie1.secure  = true
-  class Cookie < SimpleDelegator
+  class Cookie < DelegateClass(Array)
 
     # Create a new CGI::Cookie object.
     #
@@ -870,15 +872,16 @@ class CGI
     cookies = Hash.new([])
     return cookies unless raw_cookie
 
-    raw_cookie.split(/; /).each do |pairs|
+    raw_cookie.split(/[;,]\s?/).each do |pairs|
       name, values = pairs.split('=',2)
       next unless name and values
       name = CGI::unescape(name)
       values ||= ""
       values = values.split('&').collect{|v| CGI::unescape(v) }
-      unless cookies.has_key?(name)
-        cookies[name] = Cookie::new({ "name" => name, "value" => values })
+      if cookies.has_key?(name)
+        values = cookies[name].value + values
       end
+      cookies[name] = Cookie::new({ "name" => name, "value" => values })
     end
 
     cookies
@@ -1012,15 +1015,18 @@ class CGI
           end
 
           c = if bufsize < content_length
-                stdinput.read(bufsize) or ''
+                stdinput.read(bufsize)
               else
-                stdinput.read(content_length) or ''
+                stdinput.read(content_length)
               end
+          if c.nil?
+            raise EOFError, "bad content body"
+          end
           buf.concat(c)
           content_length -= c.size
         end
 
-        buf = buf.sub(/\A((?:.|\n)*?)(?:#{EOL})?#{boundary}(#{EOL}|--)/n) do
+        buf = buf.sub(/\A((?:.|\n)*?)(?:[\r\n]{1,2})?#{boundary}([\r\n]{1,2}|--)/n) do
           body.print $1
           if "--" == $2
             content_length = -1
@@ -2258,33 +2264,12 @@ class CGI
 
     extend QueryExtension
     @multipart = false
-    if "POST" != env_table['REQUEST_METHOD']
-      initialize_query()  # set @params, @cookies
+    if defined?(CGI_PARAMS)
+      warn "do not use CGI_PARAMS and CGI_COOKIES"
+      @params = CGI_PARAMS.dup
+      @cookies = CGI_COOKIES.dup
     else
-      if defined?(CGI_PARAMS)
-        @params  = CGI_PARAMS.nil?  ? nil : CGI_PARAMS.dup
-        @cookies = CGI_COOKIES.nil? ? nil : CGI_COOKIES.dup
-      else
-        initialize_query()  # set @params, @cookies
-        params  = @params.nil?  ? nil : @params.dup
-        cookies = @cookies.nil? ? nil : @cookies.dup
-        (class << self; self; end).class_eval do
-          const_set(:CGI_PARAMS,  params)
-          const_set(:CGI_COOKIES, cookies)
-        end
-        if defined?(MOD_RUBY) and (RUBY_VERSION < "1.4.3")
-          raise "Please, use ruby1.4.3 or later."
-        else
-          at_exit() do
-            if defined?(CGI_PARAMS)
-              CGI.class_eval do
-                remove_const(:CGI_PARAMS)
-                remove_const(:CGI_COOKIES)
-              end
-            end
-          end
-        end
-      end
+      initialize_query()  # set @params, @cookies
     end
     @output_cookies = nil
     @output_hidden = nil
