@@ -192,7 +192,7 @@
 #
 class OptionParser
   # :stopdoc:
-  RCSID = %w$Id: optparse.rb,v 1.40.2.7 2005/09/26 12:44:47 nobu Exp $[1..-1].each {|s| s.freeze}.freeze
+  RCSID = %w$Id: optparse.rb,v 1.40.2.10 2006/06/11 14:21:08 nobu Exp $[1..-1].each {|s| s.freeze}.freeze
   Version = (RCSID[1].split('.').collect {|s| s.to_i}.extend(Comparable).freeze if RCSID[1])
   LastModified = (Time.gm(*RCSID[2, 2].join('-').scan(/\d+/).collect {|s| s.to_i}) if RCSID[2])
   Release = RCSID[2]
@@ -521,7 +521,7 @@ class OptionParser
     #
     # See OptionParser.accept.
     #
-    def accept(t, pat = /.*/, &block)
+    def accept(t, pat = /.*/nm, &block)
       if pat
         pat.respond_to?(:match) or raise TypeError, "has no `match'"
       else
@@ -1272,7 +1272,7 @@ class OptionParser
       while arg = argv.shift
         case arg
         # long option
-        when /\A--([^=]*)(?:=(.*))?/
+        when /\A--([^=]*)(?:=(.*))?/nm
           opt, rest = $1, $2
           begin
             sw, = complete(:long, opt, true)
@@ -1287,7 +1287,7 @@ class OptionParser
           end
 
         # short option
-        when /\A-(.)((=).*|.+)?/
+        when /\A-(.)((=).*|.+)?/nm
           opt, has_arg, eq, val, rest = $1, $3, $3, $2, $2
           begin
             unless sw = search(:short, opt)
@@ -1379,6 +1379,42 @@ class OptionParser
     end
   end
 
+=begin
+--- OptionParser#getopts(argv, single_options, *long_options)
+--- OptionParser.getopts(argv, single_options, *long_options)
+    Wrapper methods for getopts.rb.
+=end #'#"#`#
+  def getopts(argv, single_options, *long_options)
+    result = {}
+
+    single_options.scan(/(.)(:)?/) do |opt, val|
+      if val
+        result[opt] = nil
+        define("-#{opt} VAL") {|val| result[opt] = val}
+      else
+        result[opt] = false
+        define("-#{opt}") {result[opt] = true}
+      end
+    end if single_options
+
+    long_options.each do |arg|
+      opt, val = arg.split(':', 2)
+      if val
+        result[opt] = val.empty? ? nil : val
+        define("--#{opt} VAL") {|val| result[opt] = val}
+      else
+        result[opt] = false
+        define("--#{opt}") {result[opt] = true}
+      end
+    end
+
+    order!(argv)
+    result
+  end
+
+  def self.getopts(*args)
+    new.getopts(*args)
+  end
 
 =begin private
 --- OptionParser#visit(id, *args) {block}
@@ -1482,7 +1518,6 @@ class OptionParser
     env = ENV[env] || ENV[env.upcase] or return
     parse(*Shellwords.shellwords(env))
   end
-
 
 =begin
 = Acceptable argument classes
@@ -1500,7 +1535,7 @@ class OptionParser
 : String
   any none-empty string, and no conversion.
 =end #'#"#`#
-  accept(String, /.+/) {|s,*|s}
+  accept(String, /.+/nm) {|s,*|s}
 
 =begin
 : Integer
@@ -1786,6 +1821,21 @@ Extends command line arguments array to parse itself.
     def permute!() options.permute!(self) end
     def parse!() options.parse!(self) end
 
+=begin
+--- OptionParser::Arguable#getopts(single_options, *long_options)
+    Substitution of getopts is possible as follows.
+
+      def getopts(*args)
+        ($OPT = ARGV.getopts(*args)).each do |opt, val|
+          eval "$OPT_#{opt.gsub(/[^A-Za-z0-9_]/, '_')} = val"
+        end
+      rescue OptionParser::ParseError
+      end
+=end #'#"#`#
+    def getopts(*args)
+      options.getopts(self, *args)
+    end
+
 =begin private
 Initializes instance variable.
 =end #'#"#`#
