@@ -2,8 +2,8 @@
 
   object.c -
 
-  $Author: matz $
-  $Date: 2005/05/11 01:11:48 $
+  $Author: akr $
+  $Date: 2006/03/02 10:46:17 $
   created at: Thu Jul 15 12:01:24 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -98,46 +98,6 @@ rb_obj_equal(obj1, obj2)
     return Qfalse;
 }
 
-
-/*
- *  Document-method: __id__
- *  Document-method: object_id
- *
- *  call-seq:
- *     obj.__id__       => fixnum
- *     obj.object_id    => fixnum
- *  
- *  Returns an integer identifier for <i>obj</i>. The same number will
- *  be returned on all calls to <code>id</code> for a given object, and
- *  no two active objects will share an id.
- *  <code>Object#object_id</code> is a different concept from the
- *  <code>:name</code> notation, which returns the symbol id of
- *  <code>name</code>. Replaces the deprecated <code>Object#id</code>.
- */
-
-
-
-/*
- *  call-seq:
- *     obj.hash    => fixnum
- *  
- *  Generates a <code>Fixnum</code> hash value for this object. This
- *  function must have the property that <code>a.eql?(b)</code> implies
- *  <code>a.hash == b.hash</code>. The hash value is used by class
- *  <code>Hash</code>. Any hash value that exceeds the capacity of a
- *  <code>Fixnum</code> will be truncated before being used.
- */
-
-VALUE
-rb_obj_id(obj)
-    VALUE obj;
-{
-    if (SPECIAL_CONST_P(obj)) {
-	return LONG2NUM((long)obj);
-    }
-    return (VALUE)((long)obj|FIXNUM_FLAG);
-}
-
 /*
  *  call-seq:
  *     obj.id    => fixnum
@@ -303,6 +263,7 @@ rb_obj_dup(obj)
     return dup;
 }
 
+/* :nodoc: */
 VALUE
 rb_obj_init_copy(obj, orig)
     VALUE obj, orig;
@@ -354,10 +315,12 @@ rb_any_to_s(obj)
     VALUE obj;
 {
     char *cname = rb_obj_classname(obj);
+    size_t len;
     VALUE str;
 
-    str = rb_str_new(0, strlen(cname)+6+16+1); /* 6:tags 16:addr 1:nul */
-    sprintf(RSTRING(str)->ptr, "#<%s:0x%lx>", cname, obj);
+    len = strlen(cname)+6+16;
+    str = rb_str_new(0, len); /* 6:tags 16:addr */
+    snprintf(RSTRING(str)->ptr, len+1, "#<%s:0x%lx>", cname, obj);
     RSTRING(str)->len = strlen(RSTRING(str)->ptr);
     if (OBJ_TAINTED(obj)) OBJ_TAINT(str);
 
@@ -433,17 +396,20 @@ rb_obj_inspect(obj)
 	&& ROBJECT(obj)->iv_tbl
 	&& ROBJECT(obj)->iv_tbl->num_entries > 0) {
 	VALUE str;
+	size_t len;
 	char *c;
 
 	c = rb_obj_classname(obj);
 	if (rb_inspecting_p(obj)) {
-	    str = rb_str_new(0, strlen(c)+10+16+1); /* 10:tags 16:addr 1:nul */
-	    sprintf(RSTRING(str)->ptr, "#<%s:0x%lx ...>", c, obj);
+	    len = strlen(c)+10+16+1;
+	    str = rb_str_new(0, len); /* 10:tags 16:addr 1:nul */
+	    snprintf(RSTRING(str)->ptr, len, "#<%s:0x%lx ...>", c, obj);
 	    RSTRING(str)->len = strlen(RSTRING(str)->ptr);
 	    return str;
 	}
-	str = rb_str_new(0, strlen(c)+6+16+1); /* 6:tags 16:addr 1:nul */
-	sprintf(RSTRING(str)->ptr, "-<%s:0x%lx", c, obj);
+	len = strlen(c)+6+16+1;
+	str = rb_str_new(0, len);     /* 6:tags 16:addr 1:nul */
+	snprintf(RSTRING(str)->ptr, len, "-<%s:0x%lx", c, obj);
 	RSTRING(str)->len = strlen(RSTRING(str)->ptr);
 	return rb_protect_inspect(inspect_obj, obj, str);
     }
@@ -1168,7 +1134,7 @@ sym_inspect(sym)
     str = rb_str_new(0, strlen(name)+1);
     RSTRING(str)->ptr[0] = ':';
     strcpy(RSTRING(str)->ptr+1, name);
-    if (rb_is_junk_id(id)) {
+    if (!rb_symname_p(name)) {
 	str = rb_str_dump(str);
 	strncpy(RSTRING(str)->ptr, ":\"", 2);
     }
@@ -2038,6 +2004,70 @@ rb_obj_ivar_set(obj, iv, val)
     return rb_ivar_set(obj, id, val);
 }
 
+/*
+ *  call-seq:
+ *     mod.class_variable_get(symbol)    => obj
+ *  
+ *  Returns the value of the given class variable (or throws a
+ *  <code>NameError</code> exception). The <code>@@</code> part of the
+ *  variable name should be included for regular class variables
+ *     
+ *     class Fred
+ *       @@foo = 99
+ *     end
+ *
+ *     def Fred.foo
+ *       class_variable_get(:@@foo)     #=> 99
+ *     end
+ */
+
+static VALUE
+rb_mod_cvar_get(obj, iv)
+    VALUE obj, iv;
+{
+    ID id = rb_to_id(iv);
+
+    if (!rb_is_class_id(id)) {
+	rb_name_error(id, "`%s' is not allowed as an class variable name", rb_id2name(id));
+    }
+    return rb_cvar_get(obj, id);
+}
+
+
+/*
+ *  call-seq:
+ *     obj.class_variable_set(symbol, obj)    => obj
+ *  
+ *  Sets the class variable names by <i>symbol</i> to
+ *  <i>object</i>.
+ *     
+ *     class Fred
+ *       @@foo = 99
+ *       def foo
+ *         @@foo
+ *       end
+ *     end
+ *
+ *     def Fred.foo
+ *       class_variable_set(:@@foo, 101)      #=> 101
+ *     end
+ *     Fred.foo
+ *     Fred.new.foo                             #=> 101
+ */
+
+static VALUE
+rb_mod_cvar_set(obj, iv, val)
+    VALUE obj, iv, val;
+{
+    ID id = rb_to_id(iv);
+
+    if (!rb_is_class_id(id)) {
+	rb_name_error(id, "`%s' is not allowed as an class variable name", rb_id2name(id));
+    }
+    rb_cvar_set(obj, id, val, Qfalse);
+    return val;
+}
+
 static VALUE
 convert_type(val, tname, method, raise)
     VALUE val;
@@ -2103,7 +2133,7 @@ rb_check_convert_type(val, type, tname, method)
 static VALUE
 rb_to_integer(val, method)
     VALUE val;
-    char *method;
+    const char *method;
 {
     VALUE v = convert_type(val, "Integer", method, Qtrue);
     if (!rb_obj_is_kind_of(v, rb_cInteger)) {
@@ -2544,10 +2574,7 @@ Init_Object()
 
     rb_define_method(rb_mKernel, "eql?", rb_obj_equal, 1);
 
-    rb_define_method(rb_mKernel, "hash", rb_obj_id, 0);
     rb_define_method(rb_mKernel, "id", rb_obj_id_obsolete, 0);
-    rb_define_method(rb_mKernel, "__id__", rb_obj_id, 0);
-    rb_define_method(rb_mKernel, "object_id", rb_obj_id, 0);
     rb_define_method(rb_mKernel, "type", rb_obj_type, 0);
     rb_define_method(rb_mKernel, "class", rb_obj_class, 0);
 
@@ -2632,7 +2659,7 @@ Init_Object()
     rb_define_method(rb_cModule, "<=", rb_class_inherited_p, 1);
     rb_define_method(rb_cModule, ">",  rb_mod_gt, 1);
     rb_define_method(rb_cModule, ">=", rb_mod_ge, 1);
-    rb_define_method(rb_cModule, "initialize_copy", rb_mod_init_copy, 1);
+    rb_define_method(rb_cModule, "initialize_copy", rb_mod_init_copy, 1); /* in class.c */
     rb_define_method(rb_cModule, "to_s", rb_mod_to_s, 0);
     rb_define_method(rb_cModule, "included_modules", 
 		     rb_mod_included_modules, 0); /* in class.c */
@@ -2668,11 +2695,13 @@ Init_Object()
 		     rb_mod_class_variables, 0); /* in variable.c */
     rb_define_private_method(rb_cModule, "remove_class_variable", 
 			     rb_mod_remove_cvar, 1); /* in variable.c */
+    rb_define_private_method(rb_cModule, "class_variable_get", rb_mod_cvar_get, 1);
+    rb_define_private_method(rb_cModule, "class_variable_set", rb_mod_cvar_set, 2);
 
     rb_define_method(rb_cClass, "allocate", rb_obj_alloc, 0);
     rb_define_method(rb_cClass, "new", rb_class_new_instance, -1);
     rb_define_method(rb_cClass, "initialize", rb_class_initialize, -1);
-    rb_define_method(rb_cClass, "initialize_copy", rb_class_init_copy, 1);
+    rb_define_method(rb_cClass, "initialize_copy", rb_class_init_copy, 1); /* in class.c */
     rb_define_method(rb_cClass, "superclass", rb_class_superclass, 0);
     rb_define_alloc_func(rb_cClass, rb_class_s_alloc);
     rb_undef_method(rb_cClass, "extend_object");
@@ -2681,8 +2710,8 @@ Init_Object()
     rb_cData = rb_define_class("Data", rb_cObject);
     rb_undef_alloc_func(rb_cData);
 
-    ruby_top_self = rb_obj_alloc(rb_cObject);
     rb_global_variable(&ruby_top_self);
+    ruby_top_self = rb_obj_alloc(rb_cObject);
     rb_define_singleton_method(ruby_top_self, "to_s", main_to_s, 0);
 
     rb_cTrueClass = rb_define_class("TrueClass", rb_cObject);
