@@ -2,8 +2,8 @@
 
   process.c -
 
-  $Author: matz $
-  $Date: 2005/02/28 02:45:19 $
+  $Author: akr $
+  $Date: 2006/06/15 06:43:29 $
   created at: Tue Aug 10 14:30:50 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -91,8 +91,8 @@ static VALUE S_Tms;
 #define HAVE_44BSD_SETGID 1
 #endif
 
-#ifdef __NetBSD__  
-#undef HAVE_SETRUID 
+#ifdef __NetBSD__
+#undef HAVE_SETRUID
 #undef HAVE_SETRGID
 #endif
 
@@ -114,10 +114,10 @@ static VALUE S_Tms;
 /*
  *  call-seq:
  *     Process.pid   => fixnum
- *  
+ *
  *  Returns the process id of this process. Not available on all
  *  platforms.
- *     
+ *
  *     Process.pid   #=> 27415
  */
 
@@ -132,15 +132,15 @@ get_pid()
 /*
  *  call-seq:
  *     Process.ppid   => fixnum
- *  
+ *
  *  Returns the process id of the parent of this process. Always
  *  returns 0 on NT. Not available on all platforms.
- *     
+ *
  *     puts "I am #{Process.pid}"
  *     Process.fork { puts "Dad is #{Process.ppid}" }
- *     
+ *
  *  <em>produces:</em>
- *     
+ *
  *     I am 27417
  *     Dad is 27417
  */
@@ -165,7 +165,7 @@ get_ppid()
  *  status of a running or terminated system process. The built-in
  *  variable <code>$?</code> is either +nil+ or a
  *  <code>Process::Status</code> object.
- *     
+ *
  *     fork { exit 99 }   #=> 26557
  *     Process.wait       #=> 26557
  *     $?.class           #=> Process::Status
@@ -174,7 +174,7 @@ get_ppid()
  *     $?.stopped?        #=> false
  *     $?.exited?         #=> true
  *     $?.exitstatus      #=> 99
- *     
+ *
  *  Posix systems record information on processes using a 16-bit
  *  integer.  The lower bits record the process status (stopped,
  *  exited, signaled) and the upper bits possibly contain additional
@@ -204,10 +204,10 @@ last_status_set(status, pid)
  *  call-seq:
  *     stat.to_i     => fixnum
  *     stat.to_int   => fixnum
- *  
+ *
  *  Returns the bits in _stat_ as a <code>Fixnum</code>. Poking
  *  around in these bits is platform dependent.
- *     
+ *
  *     fork { exit 0xab }         #=> 26566
  *     Process.wait               #=> 26566
  *     sprintf('%04x', $?.to_i)   #=> "ab00"
@@ -224,7 +224,7 @@ pst_to_i(st)
 /*
  *  call-seq:
  *     stat.to_s   => string
- *  
+ *
  *  Equivalent to _stat_<code>.to_i.to_s</code>.
  */
 
@@ -239,9 +239,9 @@ pst_to_s(st)
 /*
  *  call-seq:
  *     stat.pid   => fixnum
- *  
+ *
  *  Returns the process ID that this status object represents.
- *     
+ *
  *     fork { exit }   #=> 26569
  *     Process.wait    #=> 26569
  *     $?.pid          #=> 26569
@@ -258,7 +258,7 @@ pst_pid(st)
 /*
  *  call-seq:
  *     stat.inspect   => string
- *  
+ *
  *  Override the inspection method.
  */
 
@@ -315,7 +315,7 @@ pst_inspect(st)
 /*
  *  call-seq:
  *     stat == other   => true or false
- *  
+ *
  *  Returns +true+ if the integer value of _stat_
  *  equals <em>other</em>.
  */
@@ -332,9 +332,9 @@ pst_equal(st1, st2)
 /*
  *  call-seq:
  *     stat & num   => fixnum
- *  
+ *
  *  Logical AND of the bits in _stat_ with <em>num</em>.
- *     
+ *
  *     fork { exit 0x37 }
  *     Process.wait
  *     sprintf('%04x', $?.to_i)       #=> "3700"
@@ -354,9 +354,9 @@ pst_bitand(st1, st2)
 /*
  *  call-seq:
  *     stat >> num   => fixnum
- *  
+ *
  *  Shift the bits in _stat_ right <em>num</em> places.
- *     
+ *
  *     fork { exit 99 }   #=> 26563
  *     Process.wait       #=> 26563
  *     $?.to_i            #=> 25344
@@ -376,7 +376,7 @@ pst_rshift(st1, st2)
 /*
  *  call-seq:
  *     stat.stopped?   => true or false
- *  
+ *
  *  Returns +true+ if this process is stopped. This is only
  *  returned if the corresponding <code>wait</code> call had the
  *  <code>WUNTRACED</code> flag set.
@@ -398,7 +398,7 @@ pst_wifstopped(st)
 /*
  *  call-seq:
  *     stat.stopsig   => fixnum or nil
- *  
+ *
  *  Returns the number of the signal that caused _stat_ to stop
  *  (or +nil+ if self is not stopped).
  */
@@ -418,7 +418,7 @@ pst_wstopsig(st)
 /*
  *  call-seq:
  *     stat.signaled?   => true or false
- *  
+ *
  *  Returns +true+ if _stat_ terminated because of
  *  an uncaught signal.
  */
@@ -439,7 +439,7 @@ pst_wifsignaled(st)
 /*
  *  call-seq:
  *     stat.termsig   => fixnum or nil
- *  
+ *
  *  Returns the number of the signal that caused _stat_ to
  *  terminate (or +nil+ if self was not terminated by an
  *  uncaught signal).
@@ -460,7 +460,7 @@ pst_wtermsig(st)
 /*
  *  call-seq:
  *     stat.exited?   => true or false
- *  
+ *
  *  Returns +true+ if _stat_ exited normally (for
  *  example using an <code>exit()</code> call or finishing the
  *  program).
@@ -482,16 +482,16 @@ pst_wifexited(st)
 /*
  *  call-seq:
  *     stat.exitstatus   => fixnum or nil
- *  
+ *
  *  Returns the least significant eight bits of the return code of
  *  _stat_. Only available if <code>exited?</code> is
  *  +true+.
- *     
+ *
  *     fork { }           #=> 26572
  *     Process.wait       #=> 26572
  *     $?.exited?         #=> true
  *     $?.exitstatus      #=> 0
- *     
+ *
  *     fork { exit 99 }   #=> 26573
  *     Process.wait       #=> 26573
  *     $?.exited?         #=> true
@@ -513,7 +513,7 @@ pst_wexitstatus(st)
 /*
  *  call-seq:
  *     stat.success?   => true, false or nil
- *  
+ *
  *  Returns +true+ if _stat_ is successful, +false+ if not.
  *  Returns +nil+ if <code>exited?</code> is not +true+.
  */
@@ -533,7 +533,7 @@ pst_success_p(st)
 /*
  *  call-seq:
  *     stat.coredump?   => true or false
- *  
+ *
  *  Returns +true+ if _stat_ generated a coredump
  *  when it terminated. Not available on all platforms.
  */
@@ -679,20 +679,20 @@ waitall_each(pid, status, ary)
  *     Process.wait()                     => fixnum
  *     Process.wait(pid=-1, flags=0)      => fixnum
  *     Process.waitpid(pid=-1, flags=0)   => fixnum
- *  
+ *
  *  Waits for a child process to exit, returns its process id, and
  *  sets <code>$?</code> to a <code>Process::Status</code> object
  *  containing information on that process. Which child it waits on
  *  depends on the value of _pid_:
- * 
+ *
  *  > 0::   Waits for the child whose process ID equals _pid_.
- *			 
+ *
  *  0::     Waits for any child whose process group ID equals that of the
  *          calling process.
  *
  *  -1::    Waits for any child process (the default if no _pid_ is
  *          given).
- *			 
+ *
  *  < -1::  Waits for any child whose process group ID equals the absolute
  *          value of _pid_.
  *
@@ -701,10 +701,10 @@ waitall_each(pid, status, ary)
  *  or <code>Process::WUNTRACED</code> (return stopped children that
  *  haven't been reported). Not all flags are available on all
  *  platforms, but a flag value of zero will work on all platforms.
- *     
+ *
  *  Calling this method raises a <code>SystemError</code> if there are
  *  no child processes. Not available on all platforms.
- *     
+ *
  *     include Process
  *     fork { exit 99 }                 #=> 27429
  *     wait                             #=> 27429
@@ -751,13 +751,13 @@ proc_wait(argc, argv)
  *  call-seq:
  *     Process.wait2(pid=-1, flags=0)      => [pid, status]
  *     Process.waitpid2(pid=-1, flags=0)   => [pid, status]
- *  
+ *
  *  Waits for a child process to exit (see Process::waitpid for exact
  *  semantics) and returns an array containing the process id and the
  *  exit status (a <code>Process::Status</code> object) of that
  *  child. Raises a <code>SystemError</code> if there are no child
  *  processes.
- *     
+ *
  *     Process.fork { exit 99 }   #=> 27437
  *     pid, status = Process.wait2
  *     pid                        #=> 27437
@@ -778,11 +778,11 @@ proc_wait2(argc, argv)
 /*
  *  call-seq:
  *     Process.waitall   => [ [pid1,status1], ...]
- *  
+ *
  *  Waits for all children, returning an array of
  *  _pid_/_status_ pairs (where _status_ is a
  *  <code>Process::Status</code> object).
- *     
+ *
  *     fork { sleep 0.2; exit 2 }   #=> 27432
  *     fork { sleep 0.1; exit 1 }   #=> 27433
  *     fork {            exit 0 }   #=> 27434
@@ -813,7 +813,7 @@ proc_waitall()
 	if (pid == -1) {
 	    if (errno == ECHILD)
 		break;
-            if (errno == EINTR) {
+	    if (errno == EINTR) {
 		rb_thread_schedule();
 		continue;
 	    }
@@ -838,14 +838,14 @@ proc_waitall()
 }
 
 static VALUE
-detach_process_watcer(pid_p)
+detach_process_watcher(pid_p)
     int *pid_p;
 {
     int cpid, status;
 
     for (;;) {
 	cpid = rb_waitpid(*pid_p, &status, WNOHANG);
-	if (cpid == -1) return Qnil;
+	if (cpid != 0) return Qnil;
 	rb_thread_sleep(1);
     }
 }
@@ -854,14 +854,14 @@ VALUE
 rb_detach_process(pid)
     int pid;
 {
-    return rb_thread_create(detach_process_watcer, (void*)&pid);
+    return rb_thread_create(detach_process_watcher, (void*)&pid);
 }
 
 
 /*
  *  call-seq:
  *     Process.detach(pid)   => thread
- *  
+ *
  *  Some operating systems retain the status of terminated child
  *  processes until the parent collects that status (normally using
  *  some variant of <code>wait()</code>. If the parent never collects
@@ -872,36 +872,35 @@ rb_detach_process(pid)
  *  only when you do not intent to explicitly wait for the child to
  *  terminate.  <code>detach</code> only checks the status
  *  periodically (currently once each second).
- *     
+ *
  *  In this first example, we don't reap the first child process, so
  *  it appears as a zombie in the process status display.
- *     
+ *
  *     p1 = fork { sleep 0.1 }
  *     p2 = fork { sleep 0.2 }
  *     Process.waitpid(p2)
  *     sleep 2
  *     system("ps -ho pid,state -p #{p1}")
- *     
+ *
  *  <em>produces:</em>
- *     
+ *
  *     27389 Z
- *     
+ *
  *  In the next example, <code>Process::detach</code> is used to reap
  *  the child automatically.
- *     
+ *
  *     p1 = fork { sleep 0.1 }
  *     p2 = fork { sleep 0.2 }
  *     Process.detach(p1)
  *     Process.waitpid(p2)
  *     sleep 2
  *     system("ps -ho pid,state -p #{p1}")
- *     
+ *
  *  <em>(produces no output)</em>
  */
 
 static VALUE
-proc_detach(obj, pid)
-    VALUE pid;
+proc_detach(VALUE obj, VALUE pid)
 {
     rb_secure(2);
     return rb_detach_process(NUM2INT(pid));
@@ -1184,7 +1183,7 @@ proc_spawn(sv)
 /*
  *  call-seq:
  *     exec(command [, arg, ...])
- *  
+ *
  *  Replaces the current process by running the given external _command_.
  *  If +exec+ is given a single argument, that argument is
  *  taken as a line that is subject to shell expansion before being
@@ -1197,11 +1196,11 @@ proc_spawn(sv)
  *  subshell; otherwise, one of the <code>exec(2)</code> system calls is
  *  used, so the running command may inherit some of the environment of
  *  the original program (including open file descriptors).
- *     
+ *
  *     exec "echo *"       # echoes list of files in current directory
  *     # never get here
- *     
- *     
+ *
+ *
  *     exec "echo", "*"    # echoes an asterisk
  *     # never get here
  */
@@ -1252,11 +1251,14 @@ rb_f_exec(argc, argv)
  *  the parent, returning the process ID of the child, and once in
  *  the child, returning _nil_. The child process can exit using
  *  <code>Kernel.exit!</code> to avoid running any
- *  <code>at_exit</code> functions. The parent process should 
- *  use <code>Process.wait</code> to collect the termination statuses 
+ *  <code>at_exit</code> functions. The parent process should
+ *  use <code>Process.wait</code> to collect the termination statuses
  *  of its children or use <code>Process.detach</code> to register
  *  disinterest in their status; otherwise, the operating system
  *  may accumulate zombie processes.
+ *
+ *  The thread calling fork is the only thread in the created child process.
+ *  fork doesn't copy other threads.
  */
 
 static VALUE
@@ -1303,11 +1305,11 @@ rb_f_fork(obj)
 /*
  *  call-seq:
  *     Process.exit!(fixnum=-1)
- *  
+ *
  *  Exits the process immediately. No exit handlers are
  *  run. <em>fixnum</em> is returned to the underlying system as the
  *  exit status.
- *     
+ *
  *     Process.exit!(0)
  */
 
@@ -1351,7 +1353,13 @@ rb_syswait(pid)
     int pid;
 {
     static int overriding;
-    RETSIGTYPE (*hfunc)_((int)), (*qfunc)_((int)), (*ifunc)_((int));
+#ifdef SIGHUP
+    RETSIGTYPE (*hfunc)_((int));
+#endif
+#ifdef SIGQUIT
+    RETSIGTYPE (*qfunc)_((int));
+#endif
+    RETSIGTYPE (*ifunc)_((int));
     int status;
     int i, hooked = Qfalse;
 
@@ -1386,18 +1394,18 @@ rb_syswait(pid)
 /*
  *  call-seq:
  *     system(cmd [, arg, ...])    => true or false
- *  
+ *
  *  Executes _cmd_ in a subshell, returning +true+ if
  *  the command was found and ran successfully, +false+
  *  otherwise. An error status is available in <code>$?</code>. The
  *  arguments are processed in the same way as for
  *  <code>Kernel::exec</code>.
- *     
+ *
  *     system("echo *")
  *     system("echo", "*")
- *     
+ *
  *  <em>produces:</em>
- *     
+ *
  *     config.h main.rb
  *     *
  */
@@ -1540,14 +1548,13 @@ rb_f_system(argc, argv)
 /*
  *  call-seq:
  *     sleep([duration])    => fixnum
- *  
- *  Suspends the current thread for _duration_ seconds (which may be
- *  any number, including a +Float+ with fractional seconds). Returns the actual
- *  number of seconds slept (rounded), which may be less than that asked
- *  for if the thread was interrupted by a +SIGALRM+, or if
- *  another thread calls <code>Thread#run</code>. Zero arguments
- *  causes +sleep+ to sleep forever.
- *     
+ *
+ *  Suspends the current thread for _duration_ seconds (which may be any number,
+ *  including a +Float+ with fractional seconds). Returns the actual number of
+ *  seconds slept (rounded), which may be less than that asked for if another
+ *  thread calls <code>Thread#run</code>. Zero arguments causes +sleep+ to sleep
+ *  forever.
+ *
  *     Time.new    #=> Wed Apr 09 08:56:32 CDT 2003
  *     sleep 1.2   #=> 1
  *     Time.new    #=> Wed Apr 09 08:56:33 CDT 2003
@@ -1582,10 +1589,10 @@ rb_f_sleep(argc, argv)
 /*
  *  call-seq:
  *     Process.getpgrp   => integer
- *  
+ *
  *  Returns the process group ID for this process. Not available on
  *  all platforms.
- *     
+ *
  *     Process.getpgid(0)   #=> 25527
  *     Process.getpgrp      #=> 25527
  */
@@ -1615,7 +1622,7 @@ proc_getpgrp()
 /*
  *  call-seq:
  *     Process.setpgrp   => 0
- *  
+ *
  *  Equivalent to <code>setpgid(0,0)</code>. Not available on all
  *  platforms.
  */
@@ -1642,10 +1649,10 @@ proc_setpgrp()
 /*
  *  call-seq:
  *     Process.getpgid(pid)   => integer
- *  
+ *
  *  Returns the process group ID for the given process id. Not
  *  available on all platforms.
- *     
+ *
  *     Process.getpgid(Process.ppid())   #=> 25527
  */
 
@@ -1669,7 +1676,7 @@ proc_getpgid(obj, pid)
 /*
  *  call-seq:
  *     Process.setpgid(pid, integer)   => 0
- *  
+ *
  *  Sets the process group ID of _pid_ (0 indicates this
  *  process) to <em>integer</em>. Not available on all platforms.
  */
@@ -1696,11 +1703,11 @@ proc_setpgid(obj, pid, pgrp)
 /*
  *  call-seq:
  *     Process.setsid   => fixnum
- *  
+ *
  *  Establishes this process as a new session and process group
  *  leader, with no controlling tty. Returns the session id. Not
  *  available on all platforms.
- *     
+ *
  *     Process.setsid   #=> 27422
  */
 
@@ -1715,7 +1722,7 @@ proc_setsid()
     if (pid < 0) rb_sys_fail(0);
     return INT2FIX(pid);
 #elif defined(HAVE_SETPGRP) && defined(TIOCNOTTY)
-  pid_t pid;
+  rb_pid_t pid;
   int ret;
 
   rb_secure(2);
@@ -1744,7 +1751,7 @@ proc_setsid()
 /*
  *  call-seq:
  *     Process.getpriority(kind, integer)   => fixnum
- *  
+ *
  *  Gets the scheduling priority for specified process, process group,
  *  or user. <em>kind</em> indicates the kind of entity to find: one
  *  of <code>Process::PRIO_PGRP</code>,
@@ -1753,7 +1760,7 @@ proc_setsid()
  *  indicating the particular process, process group, or user (an id
  *  of 0 means _current_). Lower priorities are more favorable
  *  for scheduling. Not available on all platforms.
- *     
+ *
  *     Process.getpriority(Process::PRIO_USER, 0)      #=> 19
  *     Process.getpriority(Process::PRIO_PROCESS, 0)   #=> 19
  */
@@ -1782,9 +1789,9 @@ proc_getpriority(obj, which, who)
 /*
  *  call-seq:
  *     Process.setpriority(kind, integer, priority)   => 0
- *  
+ *
  *  See <code>Process#getpriority</code>.
- *     
+ *
  *     Process.setpriority(Process::PRIO_USER, 0, 19)      #=> 0
  *     Process.setpriority(Process::PRIO_PROCESS, 0, 19)   #=> 0
  *     Process.getpriority(Process::PRIO_USER, 0)          #=> 19
@@ -1811,6 +1818,115 @@ proc_setpriority(obj, which, who, prio)
 #endif
 }
 
+#if SIZEOF_RLIM_T == SIZEOF_INT
+# define RLIM2NUM(v) UINT2NUM(v)
+# define NUM2RLIM(v) NUM2UINT(v)
+#elif SIZEOF_RLIM_T == SIZEOF_LONG
+# define RLIM2NUM(v) ULONG2NUM(v)
+# define NUM2RLIM(v) NUM2ULONG(v)
+#elif SIZEOF_RLIM_T == SIZEOF_LONG_LONG
+# define RLIM2NUM(v) ULL2NUM(v)
+# define NUM2RLIM(v) NUM2ULL(v)
+#endif
+
+/*
+ *  call-seq:
+ *     Process.getrlimit(resource)   => [cur_limit, max_limit]
+ *
+ *  Gets the resource limit of the process.
+ *  _cur_limit_ means current (soft) limit and
+ *  _max_limit_ means maximum (hard) limit.
+ *
+ *  _resource_ indicates the kind of resource to limit:
+ *  such as <code>Process::RLIMIT_CORE</code>,
+ *  <code>Process::RLIMIT_CPU</code>, etc.
+ *  See Process.setrlimit for details.
+ *
+ *  _cur_limit_ and _max_limit_ may be <code>Process::RLIM_INFINITY</code>,
+ *  <code>Process::RLIM_SAVED_MAX</code> or
+ *  <code>Process::RLIM_SAVED_CUR</code>.
+ *  See Process.setrlimit and the system getrlimit(2) manual for details.
+ */
+
+static VALUE
+proc_getrlimit(VALUE obj, VALUE resource)
+{
+#if defined(HAVE_GETRLIMIT) && defined(RLIM2NUM)
+    struct rlimit rlim;
+
+    rb_secure(2);
+
+    if (getrlimit(NUM2INT(resource), &rlim) < 0) {
+        rb_sys_fail("getrlimit");
+    }
+    return rb_assoc_new(RLIM2NUM(rlim.rlim_cur), RLIM2NUM(rlim.rlim_max));
+#else
+    rb_notimplement();
+#endif
+}
+
+/*
+ *  call-seq:
+ *     Process.setrlimit(resource, cur_limit, max_limit)        => nil
+ *     Process.setrlimit(resource, cur_limit)                   => nil
+ *
+ *  Sets the resource limit of the process.
+ *  _cur_limit_ means current (soft) limit and
+ *  _max_limit_ means maximum (hard) limit.
+ *
+ *  If _max_limit_ is not given, _cur_limit_ is used.
+ *
+ *  _resource_ indicates the kind of resource to limit.
+ *  The list of resources are OS dependent.
+ *  Ruby may support following resources.
+ *
+ *  [Process::RLIMIT_CORE] core size (bytes) (SUSv3)
+ *  [Process::RLIMIT_CPU] CPU time (seconds) (SUSv3)
+ *  [Process::RLIMIT_DATA] data segment (bytes) (SUSv3)
+ *  [Process::RLIMIT_FSIZE] file size (bytes) (SUSv3)
+ *  [Process::RLIMIT_NOFILE] file descriptors (number) (SUSv3)
+ *  [Process::RLIMIT_STACK] stack size (bytes) (SUSv3)
+ *  [Process::RLIMIT_AS] total available memory (bytes) (SUSv3, NetBSD, FreeBSD, OpenBSD but 4.4BSD-Lite)
+ *  [Process::RLIMIT_MEMLOCK] total size for mlock(2) (bytes) (4.4BSD, GNU/Linux)
+ *  [Process::RLIMIT_NPROC] number of processes for the user (number) (4.4BSD, GNU/Linux)
+ *  [Process::RLIMIT_RSS] resident memory size (bytes) (4.2BSD, GNU/Linux)
+ *  [Process::RLIMIT_SBSIZE] all socket buffers (bytes) (NetBSD, FreeBSD)
+ *
+ *  Other <code>Process::RLIMIT_???</code> constants may be defined.
+ *
+ *  _cur_limit_ and _max_limit_ may be <code>Process::RLIM_INFINITY</code>,
+ *  which means that the resource is not limited.
+ *  They may be <code>Process::RLIM_SAVED_MAX</code> or
+ *  <code>Process::RLIM_SAVED_CUR</code> too.
+ *  See system setrlimit(2) manual for details.
+ *
+ */
+
+static VALUE
+proc_setrlimit(int argc, VALUE *argv, VALUE obj)
+{
+#if defined(HAVE_SETRLIMIT) && defined(NUM2RLIM)
+    VALUE resource, rlim_cur, rlim_max;
+    struct rlimit rlim;
+
+    rb_secure(2);
+
+    rb_scan_args(argc, argv, "21", &resource, &rlim_cur, &rlim_max);
+    if (rlim_max == Qnil)
+        rlim_max = rlim_cur;
+
+    rlim.rlim_cur = NUM2RLIM(rlim_cur);
+    rlim.rlim_max = NUM2RLIM(rlim_max);
+
+    if (setrlimit(NUM2INT(resource), &rlim) < 0) {
+        rb_sys_fail("setrlimit");
+    }
+    return Qnil;
+#else
+    rb_notimplement();
+#endif
+}
+
 static int under_uid_switch = 0;
 static void
 check_uid_switch()
@@ -1846,10 +1962,10 @@ check_gid_switch()
 /*
  *  call-seq:
  *     Process::Sys.setuid(integer)   => nil
- *  
+ *
  *  Set the user ID of the current process to _integer_. Not
  *  available on all platforms.
- *     
+ *
  */
 
 static VALUE
@@ -1870,10 +1986,10 @@ p_sys_setuid(obj, id)
 /*
  *  call-seq:
  *     Process::Sys.setruid(integer)   => nil
- *  
+ *
  *  Set the real user ID of the calling process to _integer_.
  *  Not available on all platforms.
- *     
+ *
  */
 
 static VALUE
@@ -1893,10 +2009,10 @@ p_sys_setruid(obj, id)
 /*
  *  call-seq:
  *     Process::Sys.seteuid(integer)   => nil
- *  
+ *
  *  Set the effective user ID of the calling process to
  *  _integer_.  Not available on all platforms.
- *     
+ *
  */
 
 static VALUE
@@ -1916,7 +2032,7 @@ p_sys_seteuid(obj, id)
 /*
  *  call-seq:
  *     Process::Sys.setreuid(rid, eid)   => nil
- *  
+ *
  *  Sets the (integer) real and/or effective user IDs of the current
  *  process to _rid_ and _eid_, respectively. A value of
  *  <code>-1</code> for either means to leave that ID unchanged. Not
@@ -1941,9 +2057,9 @@ p_sys_setreuid(obj, rid, eid)
 /*
  *  call-seq:
  *     Process::Sys.setresuid(rid, eid, sid)   => nil
- *  
+ *
  *  Sets the (integer) real, effective, and saved user IDs of the
- *  current process to _rid_, _eid_, and _sid_ respectively. A 
+ *  current process to _rid_, _eid_, and _sid_ respectively. A
  *  value of <code>-1</code> for any value means to
  *  leave that ID unchanged. Not available on all platforms.
  *
@@ -1968,9 +2084,9 @@ p_sys_setresuid(obj, rid, eid, sid)
  *     Process.uid           => fixnum
  *     Process::UID.rid      => fixnum
  *     Process::Sys.getuid   => fixnum
- *  
+ *
  *  Returns the (real) user ID of this process.
- *     
+ *
  *     Process.uid   #=> 501
  */
 
@@ -1986,7 +2102,7 @@ proc_getuid(obj)
 /*
  *  call-seq:
  *     Process.uid= integer   => numeric
- *  
+ *
  *  Sets the (integer) user ID for this process. Not available on all
  *  platforms.
  */
@@ -2021,7 +2137,7 @@ proc_setuid(obj, id)
 
 
 /********************************************************************
- * 
+ *
  * Document-class: Process::UID
  *
  *  The <code>Process::UID</code> module contains a collection of
@@ -2036,11 +2152,11 @@ static int SAVED_USER_ID;
 /*
  *  call-seq:
  *     Process::UID.change_privilege(integer)   => fixnum
- *  
+ *
  *  Change the current process's real and effective user ID to that
  *  specified by _integer_. Returns the new user ID. Not
  *  available on all platforms.
- *     
+ *
  *     [Process.uid, Process.euid]          #=> [0, 0]
  *     Process::UID.change_privilege(31)    #=> 31
  *     [Process.uid, Process.euid]          #=> [31, 31]
@@ -2050,7 +2166,6 @@ static VALUE
 p_uid_change_privilege(obj, id)
     VALUE obj, id;
 {
-    extern int errno;
     int uid;
 
     check_uid_switch();
@@ -2113,13 +2228,13 @@ p_uid_change_privilege(obj, id)
 #endif
     } else { /* unprivileged user */
 #if defined(HAVE_SETRESUID)
-	if (setresuid((getuid() == uid)? -1: uid, 
-		      (geteuid() == uid)? -1: uid, 
+	if (setresuid((getuid() == uid)? -1: uid,
+		      (geteuid() == uid)? -1: uid,
 		      (SAVED_USER_ID == uid)? -1: uid) < 0) rb_sys_fail(0);
 	SAVED_USER_ID = uid;
 #elif defined(HAVE_SETREUID) && !defined(OBSOLETE_SETREUID)
 	if (SAVED_USER_ID == uid) {
-	    if (setreuid((getuid() == uid)? -1: uid, 
+	    if (setreuid((getuid() == uid)? -1: uid,
 			 (geteuid() == uid)? -1: uid) < 0) rb_sys_fail(0);
 	} else if (getuid() != uid) {
 	    if (setreuid(uid, (geteuid() == uid)? -1: uid) < 0) rb_sys_fail(0);
@@ -2191,10 +2306,10 @@ p_uid_change_privilege(obj, id)
 /*
  *  call-seq:
  *     Process::Sys.setgid(integer)   => nil
- *  
+ *
  *  Set the group ID of the current process to _integer_. Not
  *  available on all platforms.
- *     
+ *
  */
 
 static VALUE
@@ -2214,10 +2329,10 @@ p_sys_setgid(obj, id)
 /*
  *  call-seq:
  *     Process::Sys.setrgid(integer)   => nil
- *  
+ *
  *  Set the real group ID of the calling process to _integer_.
  *  Not available on all platforms.
- *     
+ *
  */
 
 static VALUE
@@ -2238,10 +2353,10 @@ p_sys_setrgid(obj, id)
 /*
  *  call-seq:
  *     Process::Sys.setegid(integer)   => nil
- *  
+ *
  *  Set the effective group ID of the calling process to
  *  _integer_.  Not available on all platforms.
- *     
+ *
  */
 
 static VALUE
@@ -2261,7 +2376,7 @@ p_sys_setegid(obj, id)
 /*
  *  call-seq:
  *     Process::Sys.setregid(rid, eid)   => nil
- *  
+ *
  *  Sets the (integer) real and/or effective group IDs of the current
  *  process to <em>rid</em> and <em>eid</em>, respectively. A value of
  *  <code>-1</code> for either means to leave that ID unchanged. Not
@@ -2285,7 +2400,7 @@ p_sys_setregid(obj, rid, eid)
 /*
  *  call-seq:
  *     Process::Sys.setresgid(rid, eid, sid)   => nil
- *  
+ *
  *  Sets the (integer) real, effective, and saved user IDs of the
  *  current process to <em>rid</em>, <em>eid</em>, and <em>sid</em>
  *  respectively. A value of <code>-1</code> for any value means to
@@ -2310,7 +2425,7 @@ p_sys_setresgid(obj, rid, eid, sid)
 /*
  *  call-seq:
  *     Process::Sys.issetugid   => true or false
- *  
+ *
  *  Returns +true+ if the process was created as a result
  *  of an execve(2) system call which had either of the setuid or
  *  setgid bits set (and extra privileges were given as a result) or
@@ -2342,9 +2457,9 @@ p_sys_issetugid(obj)
  *     Process.gid           => fixnum
  *     Process::GID.rid      => fixnum
  *     Process::Sys.getgid   => fixnum
- *  
+ *
  *  Returns the (real) group ID for this process.
- *     
+ *
  *     Process.gid   #=> 500
  */
 
@@ -2360,7 +2475,7 @@ proc_getgid(obj)
 /*
  *  call-seq:
  *     Process.gid= fixnum   => fixnum
- *  
+ *
  *  Sets the group ID for this process.
  */
 
@@ -2399,7 +2514,7 @@ static size_t maxgroups = 32;
 /*
  *  call-seq:
  *     Process.groups   => array
- *  
+ *
  *  Get an <code>Array</code> of the gids of groups in the
  *  supplemental group access list for this process.
  *
@@ -2413,18 +2528,18 @@ proc_getgroups(VALUE obj)
 #ifdef HAVE_GETGROUPS
     VALUE ary;
     size_t ngroups;
-    gid_t *groups;
+    rb_gid_t *groups;
     int i;
 
-    groups = ALLOCA_N(gid_t, maxgroups);
+    groups = ALLOCA_N(rb_gid_t, maxgroups);
 
     ngroups = getgroups(maxgroups, groups);
     if (ngroups == -1)
-        rb_sys_fail(0);
+	rb_sys_fail(0);
 
     ary = rb_ary_new();
     for (i = 0; i < ngroups; i++)
-        rb_ary_push(ary, INT2NUM(groups[i]));
+	rb_ary_push(ary, INT2NUM(groups[i]));
 
     return ary;
 #else
@@ -2437,7 +2552,7 @@ proc_getgroups(VALUE obj)
 /*
  *  call-seq:
  *     Process.groups= array   => array
- *  
+ *
  *  Set the supplemental group access list to the given
  *  <code>Array</code> of group IDs.
  *
@@ -2452,7 +2567,7 @@ proc_setgroups(VALUE obj, VALUE ary)
 {
 #ifdef HAVE_SETGROUPS
     size_t ngroups;
-    gid_t *groups;
+    rb_gid_t *groups;
     int i;
     struct group *gr;
 
@@ -2460,15 +2575,15 @@ proc_setgroups(VALUE obj, VALUE ary)
 
     ngroups = RARRAY(ary)->len;
     if (ngroups > maxgroups)
-        rb_raise(rb_eArgError, "too many groups, %d max", maxgroups);
+	rb_raise(rb_eArgError, "too many groups, %d max", maxgroups);
 
-    groups = ALLOCA_N(gid_t, ngroups);
+    groups = ALLOCA_N(rb_gid_t, ngroups);
 
     for (i = 0; i < ngroups && i < RARRAY(ary)->len; i++) {
-        VALUE g = RARRAY(ary)->ptr[i];
+	VALUE g = RARRAY(ary)->ptr[i];
 
 	if (FIXNUM_P(g)) {
-            groups[i] = FIX2INT(g);
+	    groups[i] = FIX2INT(g);
 	}
 	else {
 	    VALUE tmp = rb_check_string_type(g);
@@ -2479,7 +2594,7 @@ proc_setgroups(VALUE obj, VALUE ary)
 	    else {
 		gr = getgrnam(RSTRING(tmp)->ptr);
 		if (gr == NULL)
-		    rb_raise(rb_eArgError, 
+		    rb_raise(rb_eArgError,
 			     "can't find group for %s", RSTRING(tmp)->ptr);
 		groups[i] = gr->gr_gid;
 	    }
@@ -2488,7 +2603,7 @@ proc_setgroups(VALUE obj, VALUE ary)
 
     i = setgroups(ngroups, groups);
     if (i == -1)
-        rb_sys_fail(0);
+	rb_sys_fail(0);
 
     return proc_getgroups(obj);
 #else
@@ -2501,7 +2616,7 @@ proc_setgroups(VALUE obj, VALUE ary)
 /*
  *  call-seq:
  *     Process.initgroups(username, gid)   => array
- *  
+ *
  *  Initializes the supplemental group access list by reading the
  *  system group database and using all groups of which the given user
  *  is a member. The group with the specified <em>gid</em> is also
@@ -2520,7 +2635,7 @@ proc_initgroups(obj, uname, base_grp)
     VALUE obj, uname, base_grp;
 {
 #ifdef HAVE_INITGROUPS
-    if (initgroups(StringValuePtr(uname), (gid_t)NUM2INT(base_grp)) != 0) {
+    if (initgroups(StringValuePtr(uname), (rb_gid_t)NUM2INT(base_grp)) != 0) {
 	rb_sys_fail(0);
     }
     return proc_getgroups(obj);
@@ -2534,10 +2649,10 @@ proc_initgroups(obj, uname, base_grp)
 /*
  *  call-seq:
  *     Process.maxgroups   => fixnum
- *  
+ *
  *  Returns the maximum number of gids allowed in the supplemental
  *  group access list.
- *     
+ *
  *     Process.maxgroups   #=> 32
  */
 
@@ -2552,14 +2667,13 @@ proc_getmaxgroups(obj)
 /*
  *  call-seq:
  *     Process.maxgroups= fixnum   => fixnum
- *  
+ *
  *  Sets the maximum number of gids allowed in the supplemental group
  *  access list.
  */
 
 static VALUE
-proc_setmaxgroups(obj, val)
-    VALUE obj;
+proc_setmaxgroups(VALUE obj, VALUE val)
 {
     size_t  ngroups = FIX2INT(val);
 
@@ -2573,7 +2687,7 @@ proc_setmaxgroups(obj, val)
 
 
 /********************************************************************
- * 
+ *
  * Document-class: Process::GID
  *
  *  The <code>Process::GID</code> module contains a collection of
@@ -2588,7 +2702,7 @@ static int SAVED_GROUP_ID;
 /*
  *  call-seq:
  *     Process::GID.change_privilege(integer)   => fixnum
- *  
+ *
  *  Change the current process's real and effective group ID to that
  *  specified by _integer_. Returns the new group ID. Not
  *  available on all platforms.
@@ -2602,7 +2716,6 @@ static VALUE
 p_gid_change_privilege(obj, id)
     VALUE obj, id;
 {
-    extern int errno;
     int gid;
 
     check_gid_switch();
@@ -2666,13 +2779,13 @@ p_gid_change_privilege(obj, id)
 #endif
     } else { /* unprivileged user */
 #if defined(HAVE_SETRESGID)
-	if (setresgid((getgid() == gid)? -1: gid, 
-		      (getegid() == gid)? -1: gid, 
+	if (setresgid((getgid() == gid)? -1: gid,
+		      (getegid() == gid)? -1: gid,
 		      (SAVED_GROUP_ID == gid)? -1: gid) < 0) rb_sys_fail(0);
 	SAVED_GROUP_ID = gid;
 #elif defined(HAVE_SETREGID) && !defined(OBSOLETE_SETREGID)
 	if (SAVED_GROUP_ID == gid) {
-	    if (setregid((getgid() == gid)? -1: gid, 
+	    if (setregid((getgid() == gid)? -1: gid,
 			 (getegid() == gid)? -1: gid) < 0) rb_sys_fail(0);
 	} else if (getgid() != gid) {
 	    if (setregid(gid, (getegid() == gid)? -1: gid) < 0) rb_sys_fail(0);
@@ -2745,9 +2858,9 @@ p_gid_change_privilege(obj, id)
  *     Process.euid           => fixnum
  *     Process::UID.eid       => fixnum
  *     Process::Sys.geteuid   => fixnum
- *  
+ *
  *  Returns the effective user ID for this process.
- *     
+ *
  *     Process.euid   #=> 501
  */
 
@@ -2763,7 +2876,7 @@ proc_geteuid(obj)
 /*
  *  call-seq:
  *     Process.euid= integer
- *  
+ *
  *  Sets the effective user ID for this process. Not available on all
  *  platforms.
  */
@@ -2833,11 +2946,11 @@ rb_seteuid_core(euid)
  *  call-seq:
  *     Process::UID.grant_privilege(integer)   => fixnum
  *     Process::UID.eid= integer               => fixnum
- *  
+ *
  *  Set the effective user ID, and if possible, the saved user ID of
  *  the process to the given _integer_. Returns the new
  *  effective user ID. Not available on all platforms.
- *     
+ *
  *     [Process.uid, Process.euid]          #=> [0, 0]
  *     Process::UID.grant_privilege(31)     #=> 31
  *     [Process.uid, Process.euid]          #=> [0, 31]
@@ -2856,10 +2969,10 @@ p_uid_grant_privilege(obj, id)
  *     Process.egid          => fixnum
  *     Process::GID.eid      => fixnum
  *     Process::Sys.geteid   => fixnum
- *  
+ *
  *  Returns the effective group ID for this process. Not available on
  *  all platforms.
- *     
+ *
  *     Process.egid   #=> 500
  */
 
@@ -2876,7 +2989,7 @@ proc_getegid(obj)
 /*
  *  call-seq:
  *     Process.egid = fixnum   => fixnum
- *  
+ *
  *  Sets the effective group ID for this process. Not available on all
  *  platforms.
  */
@@ -2947,11 +3060,11 @@ rb_setegid_core(egid)
  *  call-seq:
  *     Process::GID.grant_privilege(integer)    => fixnum
  *     Process::GID.eid = integer               => fixnum
- *  
+ *
  *  Set the effective group ID, and if possible, the saved group ID of
  *  the process to the given _integer_. Returns the new
  *  effective group ID. Not available on all platforms.
- *     
+ *
  *     [Process.gid, Process.egid]          #=> [0, 0]
  *     Process::GID.grant_privilege(31)     #=> 33
  *     [Process.gid, Process.egid]          #=> [0, 33]
@@ -2968,10 +3081,10 @@ p_gid_grant_privilege(obj, id)
 /*
  *  call-seq:
  *     Process::UID.re_exchangeable?   => true or false
- *  
+ *
  *  Returns +true+ if the real and effective user IDs of a
  *  process may be exchanged on the current platform.
- *     
+ *
  */
 
 static VALUE
@@ -2990,10 +3103,10 @@ p_uid_exchangeable()
 /*
  *  call-seq:
  *     Process::UID.re_exchange   => fixnum
- *  
+ *
  *  Exchange real and effective user IDs and return the new effective
  *  user ID. Not available on all platforms.
- *     
+ *
  *     [Process.uid, Process.euid]   #=> [0, 31]
  *     Process::UID.re_exchange      #=> 0
  *     [Process.uid, Process.euid]   #=> [31, 0]
@@ -3026,10 +3139,10 @@ p_uid_exchange(obj)
 /*
  *  call-seq:
  *     Process::GID.re_exchangeable?   => true or false
- *  
+ *
  *  Returns +true+ if the real and effective group IDs of a
  *  process may be exchanged on the current platform.
- *     
+ *
  */
 
 static VALUE
@@ -3048,10 +3161,10 @@ p_gid_exchangeable()
 /*
  *  call-seq:
  *     Process::GID.re_exchange   => fixnum
- *  
+ *
  *  Exchange real and effective group IDs and return the new effective
  *  group ID. Not available on all platforms.
- *     
+ *
  *     [Process.gid, Process.egid]   #=> [0, 33]
  *     Process::GID.re_exchange      #=> 0
  *     [Process.gid, Process.egid]   #=> [33, 0]
@@ -3085,10 +3198,10 @@ p_gid_exchange(obj)
 /*
  *  call-seq:
  *     Process::UID.sid_available?   => true or false
- *  
+ *
  *  Returns +true+ if the current platform has saved user
  *  ID functionality.
- *     
+ *
  */
 
 static VALUE
@@ -3116,20 +3229,19 @@ p_uid_sw_ensure(id)
  *  call-seq:
  *     Process::UID.switch              => fixnum
  *     Process::UID.switch {|| block}   => object
- *  
+ *
  *  Switch the effective and real user IDs of the current process. If
  *  a <em>block</em> is given, the user IDs will be switched back
  *  after the block is executed. Returns the new effective user ID if
  *  called without a block, and the return value of the block if one
  *  is given.
- *     
+ *
  */
 
 static VALUE
 p_uid_switch(obj)
     VALUE obj;
 {
-    extern int errno;
     int uid, euid;
 
     check_uid_switch();
@@ -3171,7 +3283,6 @@ static VALUE
 p_uid_switch(obj)
     VALUE obj;
 {
-    extern int errno;
     int uid, euid;
 
     check_uid_switch();
@@ -3199,10 +3310,10 @@ p_uid_switch(obj)
 /*
  *  call-seq:
  *     Process::GID.sid_available?   => true or false
- *  
+ *
  *  Returns +true+ if the current platform has saved group
  *  ID functionality.
- *     
+ *
  */
 
 static VALUE
@@ -3229,20 +3340,19 @@ p_gid_sw_ensure(id)
  *  call-seq:
  *     Process::GID.switch              => fixnum
  *     Process::GID.switch {|| block}   => object
- *  
+ *
  *  Switch the effective and real group IDs of the current process. If
  *  a <em>block</em> is given, the group IDs will be switched back
  *  after the block is executed. Returns the new effective group ID if
  *  called without a block, and the return value of the block if one
  *  is given.
- *     
+ *
  */
 
 static VALUE
 p_gid_switch(obj)
     VALUE obj;
 {
-    extern int errno;
     int gid, egid;
 
     check_gid_switch();
@@ -3283,7 +3393,6 @@ static VALUE
 p_gid_switch(obj)
     VALUE obj;
 {
-    extern int errno;
     int gid, egid;
 
     check_gid_switch();
@@ -3309,11 +3418,11 @@ p_gid_switch(obj)
 /*
  *  call-seq:
  *     Process.times   => aStructTms
- *  
+ *
  *  Returns a <code>Tms</code> structure (see <code>Struct::Tms</code>
  *  on page 388) that contains user and system CPU times for this
  *  process.
- *     
+ *
  *     t = Process.times
  *     [ t.utime, t.stime ]   #=> [0.0, 0.02]
  */
@@ -3435,6 +3544,53 @@ Init_process()
     rb_define_const(rb_mProcess, "PRIO_USER", INT2FIX(PRIO_USER));
 #endif
 
+    rb_define_module_function(rb_mProcess, "getrlimit", proc_getrlimit, 1);
+    rb_define_module_function(rb_mProcess, "setrlimit", proc_setrlimit, -1);
+#ifdef RLIM2NUM
+#ifdef RLIM_INFINITY
+    rb_define_const(rb_mProcess, "RLIM_INFINITY", RLIM2NUM(RLIM_INFINITY));
+#endif
+#ifdef RLIM_SAVED_MAX
+    rb_define_const(rb_mProcess, "RLIM_SAVED_MAX", RLIM2NUM(RLIM_SAVED_MAX));
+#endif
+#ifdef RLIM_SAVED_CUR
+    rb_define_const(rb_mProcess, "RLIM_SAVED_CUR", RLIM2NUM(RLIM_SAVED_CUR));
+#endif
+#ifdef RLIMIT_CORE
+    rb_define_const(rb_mProcess, "RLIMIT_CORE", INT2FIX(RLIMIT_CORE));
+#endif
+#ifdef RLIMIT_CPU
+    rb_define_const(rb_mProcess, "RLIMIT_CPU", INT2FIX(RLIMIT_CPU));
+#endif
+#ifdef RLIMIT_DATA
+    rb_define_const(rb_mProcess, "RLIMIT_DATA", INT2FIX(RLIMIT_DATA));
+#endif
+#ifdef RLIMIT_FSIZE
+    rb_define_const(rb_mProcess, "RLIMIT_FSIZE", INT2FIX(RLIMIT_FSIZE));
+#endif
+#ifdef RLIMIT_NOFILE
+    rb_define_const(rb_mProcess, "RLIMIT_NOFILE", INT2FIX(RLIMIT_NOFILE));
+#endif
+#ifdef RLIMIT_STACK
+    rb_define_const(rb_mProcess, "RLIMIT_STACK", INT2FIX(RLIMIT_STACK));
+#endif
+#ifdef RLIMIT_AS
+    rb_define_const(rb_mProcess, "RLIMIT_AS", INT2FIX(RLIMIT_AS));
+#endif
+#ifdef RLIMIT_MEMLOCK
+    rb_define_const(rb_mProcess, "RLIMIT_MEMLOCK", INT2FIX(RLIMIT_MEMLOCK));
+#endif
+#ifdef RLIMIT_NPROC
+    rb_define_const(rb_mProcess, "RLIMIT_NPROC", INT2FIX(RLIMIT_NPROC));
+#endif
+#ifdef RLIMIT_RSS
+    rb_define_const(rb_mProcess, "RLIMIT_RSS", INT2FIX(RLIMIT_RSS));
+#endif
+#ifdef RLIMIT_SBSIZE
+    rb_define_const(rb_mProcess, "RLIMIT_SBSIZE", INT2FIX(RLIMIT_SBSIZE));
+#endif
+#endif
+
     rb_define_module_function(rb_mProcess, "uid", proc_getuid, 0);
     rb_define_module_function(rb_mProcess, "uid=", proc_setuid, 1);
     rb_define_module_function(rb_mProcess, "gid", proc_getgid, 0);
