@@ -2,7 +2,7 @@
 
   re.c -
 
-  $Author: nobu $
+  $Author: matz $
   created at: Mon Aug  9 18:24:49 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -70,10 +70,11 @@ static const char casetable[] = {
 #endif
 
 int
-rb_memcicmp(p1, p2, len)
-    char *p1, *p2;
+rb_memcicmp(x, y, len)
+    const void *x, *y;
     long len;
 {
+    const unsigned char *p1 = x, *p2 = y;
     int tmp;
 
     while (len--) {
@@ -85,7 +86,7 @@ rb_memcicmp(p1, p2, len)
 
 int
 rb_memcmp(p1, p2, len)
-    char *p1, *p2;
+    const void *p1, *p2;
     long len;
 {
     if (!ruby_ignorecase) {
@@ -96,16 +97,16 @@ rb_memcmp(p1, p2, len)
 
 long
 rb_memsearch(x0, m, y0, n)
-    char *x0, *y0;
+    const void *x0, *y0;
     long m, n;
 {
-    unsigned char *x = (unsigned char *)x0, *y = (unsigned char *)y0;
-    unsigned char *s, *e;
+    const unsigned char *x = (unsigned char *)x0, *y = (unsigned char *)y0;
+    const unsigned char *s, *e;
     long i;
     int d;
     unsigned long hx, hy;
 
-#define KR_REHASH(a, b, h) (((h) << 1) - ((a)<<d) + (b))
+#define KR_REHASH(a, b, h) (((h) << 1) - (((unsigned long)(a))<<d) + (b))
 
     if (m > n) return -1;
     s = y; e = s + n - m;
@@ -151,6 +152,7 @@ rb_memsearch(x0, m, y0, n)
     return s-y;
 }
 
+#define REG_LITERAL FL_USER5
 #define REG_CASESTATE  FL_USER0
 #define KCODE_NONE  0
 #define KCODE_EUC   FL_USER1
@@ -1332,6 +1334,11 @@ rb_reg_initialize(obj, s, len, options)
 {
     struct RRegexp *re = RREGEXP(obj);
 
+    if (!OBJ_TAINTED(obj) && rb_safe_level() >= 4)
+	rb_raise(rb_eSecurityError, "Insecure: can't modify regexp");
+    rb_check_frozen(obj);
+    if (FL_TEST(obj, REG_LITERAL))
+	rb_raise(rb_eSecurityError, "can't modify literal regexp");
     if (re->ptr) re_free_pattern(re->ptr);
     if (re->str) free(re->str);
     re->ptr = 0;
@@ -1371,6 +1378,7 @@ rb_reg_initialize(obj, s, len, options)
     if (options & ~0xf) {
 	kcode_reset_option();
     }
+    if (ruby_in_compile) FL_SET(obj, REG_LITERAL);
 }
 
 static VALUE rb_reg_s_alloc _((VALUE));
@@ -1408,6 +1416,7 @@ VALUE
 rb_reg_regcomp(str)
     VALUE str;
 {
+    volatile VALUE save_str = str;
     if (reg_cache && RREGEXP(reg_cache)->len == RSTRING(str)->len
 	&& case_cache == ruby_ignorecase
 	&& kcode_cache == reg_kcode
@@ -1652,7 +1661,6 @@ rb_reg_initialize_m(argc, argv, self)
     long len;
     int flags = 0;
 
-    rb_check_frozen(self);
     if (argc == 0 || argc > 3) {
 	rb_raise(rb_eArgError, "wrong number of arguments");
     }
