@@ -2,8 +2,8 @@
 
   signal.c -
 
-  $Author: matz $
-  $Date: 2005/02/28 02:45:19 $
+  $Author: nagai $
+  $Date: 2006/06/23 21:37:08 $
   created at: Tue Dec 20 10:13:44 JST 1994
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -33,6 +33,7 @@ static struct signals {
     char *signm;
     int  signo;
 } siglist [] = {
+    {"EXIT", 0},
 #ifdef SIGHUP
     {"HUP", SIGHUP},
 #endif
@@ -508,7 +509,7 @@ static RETSIGTYPE
 sigexit(sig)
     int sig;
 {
-    rb_exit(0);
+    rb_thread_signal_exit();
 }
 
 static VALUE
@@ -577,7 +578,7 @@ trap(arg)
 	    rb_raise(rb_eArgError, "unsupported signal SIG%s", s);
     }
 
-    if (sig < 0 || sig > NSIG) {
+    if (sig < 0 || sig >= NSIG) {
 	rb_raise(rb_eArgError, "invalid signal number (%d)", sig);
     }
 #if defined(HAVE_SETITIMER)
