@@ -47,12 +47,6 @@
 #define bool int
 #endif
 
-#ifdef _M_IX86
-# define WIN95 1
-#else
-# undef  WIN95
-#endif
-
 #if defined __BORLANDC__ || defined _WIN32_WCE
 #  define _filbuf _fgetc
 #  define _flsbuf _fputc
@@ -182,7 +176,13 @@ static struct {
     {	ERROR_NESTING_NOT_ALLOWED,	EAGAIN		},
     {	ERROR_NOT_ENOUGH_QUOTA,		ENOMEM		},
     {	WSAENAMETOOLONG,		ENAMETOOLONG	},
-    {	WSAENOTEMPTY,			ENOTEMPTY	}
+    {	WSAENOTEMPTY,			ENOTEMPTY	},
+    {	WSAEINTR,			EINTR		},
+    {	WSAEBADF,			EBADF		},
+    {	WSAEACCES,			EACCES		},
+    {	WSAEFAULT,			EFAULT		},
+    {	WSAEINVAL,			EINVAL		},
+    {	WSAEMFILE,			EMFILE		},
 };
 
 static int
@@ -209,10 +209,10 @@ map_errno(DWORD winerr)
 static char *NTLoginName;
 
 #ifdef WIN95
-DWORD Win32System = (DWORD)-1;
+static DWORD Win32System = (DWORD)-1;
 
-static DWORD
-IdOS(void)
+DWORD
+rb_w32_osid(void)
 {
     static OSVERSIONINFO osver;
 
@@ -224,21 +224,11 @@ IdOS(void)
     }
     return (Win32System);
 }
-
-static int 
-IsWin95(void) {
-    return (IdOS() == VER_PLATFORM_WIN32_WINDOWS);
-}
-
-static int
-IsWinNT(void) {
-    return (IdOS() == VER_PLATFORM_WIN32_NT);
-}
-#else
-# define IsWinNT() TRUE
-# define IsWin95() FALSE
 #endif
 
+#define IsWinNT() rb_w32_iswinnt()
+#define IsWin95() rb_w32_iswin95()
+
 /* main thread constants */
 static struct {
     HANDLE handle;
@@ -266,7 +256,20 @@ GetCurrentThreadHandle(void)
 /* simulate flock by locking a range on the file */
 
 
-#define LK_ERR(f,i) ((f) ? (i = 0) : (errno = GetLastError() == ERROR_LOCK_VIOLATION ? EWOULDBLOCK : EACCES))
+#define LK_ERR(f,i) \
+    do {								\
+	if (f)								\
+	    i = 0;							\
+	else {								\
+	    DWORD err = GetLastError();					\
+	    if (err == ERROR_LOCK_VIOLATION)				\
+		errno = EWOULDBLOCK;					\
+	    else if (err == ERROR_NOT_LOCKED)				\
+		i = 0;							\
+	    else							\
+		errno = map_errno(err);					\
+	}								\
+    } while (0)
 #define LK_LEN      ULONG_MAX
 
 static VALUE
@@ -357,6 +360,13 @@ flock(int fd, int oper)
 			      (DWORD)-1);
 }
 
+#if _MSC_VER >= 1400
+static void invalid_parameter(const wchar_t *expr, const wchar_t *func, const wchar_t *file, unsigned int line, uintptr_t dummy)
+{
+    // nothing to do
+}
+#endif
+
 //
 // Initialization stuff
 //
@@ -367,6 +377,10 @@ NtInitialize(int *argc, char ***argv)
     WORD version;
     int ret;
 
+#if _MSC_VER >= 1400
+    _set_invalid_parameter_handler(invalid_parameter);
+#endif
+
     //
     // subvert cmd.exe's feeble attempt at command line parsing
     //
@@ -1177,7 +1191,7 @@ rb_w32_cmdvector(const char *cmd, char ***vec)
 
 	    switch (*ptr) {
 	      case '\\':
-		slashes++;
+		if (quote != '\'') slashes++;
 	        break;
 
 	      case ' ':
@@ -1256,7 +1270,7 @@ rb_w32_cmdvector(const char *cmd, char ***vec)
 		switch (*p) {
 		  case '\\':
 		    p++;
-		    slashes++;
+		    if (quote != '\'') slashes++;
 		    break;
 
 		  case '\'':
@@ -1464,6 +1478,20 @@ rb_w32_opendir(const char *filename)
     return p;
 }
 
+//
+// Move to next entry
+//
+
+static void
+move_to_next_entry(DIR *dirp)
+{
+    if (dirp->curr) {
+	dirp->curr += strlen(dirp->curr) + 1;
+	if (dirp->curr >= (dirp->start + dirp->size)) {
+	    dirp->curr = NULL;
+	}
+    }
+}
 
 //
 // Readdir just returns the current string pointer and bumps the
@@ -1473,7 +1501,6 @@ rb_w32_opendir(const char *filename)
 struct direct  *
 rb_w32_readdir(DIR *dirp)
 {
-    int         len;
     static int  dummy = 0;
 
     if (dirp->curr) {
@@ -1482,9 +1509,8 @@ rb_w32_readdir(DIR *dirp)
 	// first set up the structure to return
 	//
 
-	len = strlen(dirp->curr);
 	strcpy(dirp->dirstr.d_name, dirp->curr);
-	dirp->dirstr.d_namlen = len;
+	dirp->dirstr.d_namlen = strlen(dirp->curr);
 
 	//
 	// Fake inode
@@ -1495,10 +1521,7 @@ rb_w32_readdir(DIR *dirp)
 	// Now set up for the next call to readdir
 	//
 
-	dirp->curr += len + 1;
-	if (dirp->curr >= (dirp->start + dirp->size)) {
-	    dirp->curr = NULL;
-	}
+	move_to_next_entry(dirp);
 
 	return &(dirp->dirstr);
 
@@ -1513,7 +1536,15 @@ rb_w32_readdir(DIR *dirp)
 long
 rb_w32_telldir(DIR *dirp)
 {
-	return (long) dirp->curr;	/* ouch! pointer to long cast */
+    long loc = 0; char *p = dirp->curr;
+
+    rb_w32_rewinddir(dirp);
+
+    while (p != dirp->curr) {
+	move_to_next_entry(dirp); loc++;
+    }
+
+    return loc;
 }
 
 //
@@ -1523,7 +1554,11 @@ rb_w32_telldir(DIR *dirp)
 void
 rb_w32_seekdir(DIR *dirp, long loc)
 {
-	dirp->curr = (char *) loc;	/* ouch! long to pointer cast */
+    rb_w32_rewinddir(dirp);
+
+    while (dirp->curr && loc-- > 0) {
+	move_to_next_entry(dirp);
+    }
 }
 
 //
@@ -1533,7 +1568,7 @@ rb_w32_seekdir(DIR *dirp, long loc)
 void
 rb_w32_rewinddir(DIR *dirp)
 {
-	dirp->curr = dirp->start;
+    dirp->curr = dirp->start;
 }
 
 //
@@ -1543,8 +1578,8 @@ rb_w32_rewinddir(DIR *dirp)
 void
 rb_w32_closedir(DIR *dirp)
 {
-	free(dirp->start);
-	free(dirp);
+    free(dirp->start);
+    free(dirp);
 }
 
 EXTERN_C void __cdecl _lock_fhandle(int);
@@ -1572,6 +1607,10 @@ typedef struct	{
 #ifdef MSVCRT_THREADS
     int lockinitflag;
     CRITICAL_SECTION lock;
+#if _MSC_VER >= 1400
+    char textmode;
+    char pipech2[2];
+#endif
 #endif
 }	ioinfo;
 
@@ -1728,9 +1767,9 @@ rb_w32_strerror(int e)
 	buffer[sizeof(buffer) - 1] = 0;
     }
 
-    for (p = buffer + strlen(buffer) - 1; buffer <= p; p--) {
-	if (*p != '\r' && *p != '\n') break;
-	*p = 0;
+    p = buffer;
+    while ((p = strpbrk(p, "\r\n")) != NULL) {
+	memmove(p, p + 1, strlen(p));
     }
     return buffer;
 }
@@ -1956,12 +1995,21 @@ StartSockets ()
     WORD version;
     WSADATA retdata;
     int ret;
+#ifndef USE_WINSOCK2
     int iSockOpt;
-    
+#endif
+
     //
     // initalize the winsock interface and insure that it's
     // cleaned up at exit.
     //
+#ifdef USE_WINSOCK2
+    version = MAKEWORD(2, 0);
+    if (WSAStartup(version, &retdata))
+	rb_fatal ("Unable to locate winsock library!\n");
+    if (LOBYTE(retdata.wVersion) != 2)
+	rb_fatal("could not find version 2 of winsock dll\n");
+#else
     version = MAKEWORD(1, 1);
     if (ret = WSAStartup(version, &retdata))
 	rb_fatal ("Unable to locate winsock library!\n");
@@ -1970,23 +2018,26 @@ StartSockets ()
 
     if (HIBYTE(retdata.wVersion) != 1)
 	rb_fatal("could not find version 1 of winsock dll\n");
+#endif	/* USE_WINSOCK2 */
 
     atexit((void (*)(void)) WSACleanup);
 
-#ifndef SO_SYNCHRONOUS_NONALERT
-#define SO_SYNCHRONOUS_NONALERT 0x20
-#endif
+#ifndef USE_WINSOCK2
+# ifndef SO_SYNCHRONOUS_NONALERT
+#  define SO_SYNCHRONOUS_NONALERT 0x20
+# endif
 
     iSockOpt = SO_SYNCHRONOUS_NONALERT;
     /*
      * Enable the use of sockets as filehandles
      */
-#ifndef SO_OPENTYPE
-#define SO_OPENTYPE     0x7008
-#endif
+# ifndef SO_OPENTYPE
+#  define SO_OPENTYPE     0x7008
+# endif
 
     setsockopt(INVALID_SOCKET, SOL_SOCKET, SO_OPENTYPE,
 	       (char *)&iSockOpt, sizeof(iSockOpt));
+#endif	/* USE_WINSOCK2 */
 
     main_thread.handle = GetCurrentThreadHandle();
     main_thread.id = GetCurrentThreadId();
@@ -2251,7 +2302,54 @@ rb_w32_shutdown(int s, int how)
     return r;
 }
 
+#ifdef USE_WINSOCK2
+static SOCKET
+open_ifs_socket(int af, int type, int protocol)
+{
+    unsigned long proto_buffers_len = 0;
+    int error_code;
+    SOCKET out = INVALID_SOCKET;
+
+    if (WSAEnumProtocols(NULL, NULL, &proto_buffers_len) == SOCKET_ERROR) {
+	error_code = WSAGetLastError();
+	if (error_code == WSAENOBUFS) {
+	    WSAPROTOCOL_INFO *proto_buffers;
+	    int protocols_available = 0;
+
+	    proto_buffers = (WSAPROTOCOL_INFO *)malloc(proto_buffers_len);
+
+	    protocols_available =
+		WSAEnumProtocols(NULL, proto_buffers, &proto_buffers_len);
+	    if (protocols_available != SOCKET_ERROR) {
+		int i;
+		for (i = 0; i < protocols_available; i++) {
+		    if ((af != AF_UNSPEC && af != proto_buffers[i].iAddressFamily) ||
+			(type != proto_buffers[i].iSocketType) ||
+			(protocol != 0 && protocol != proto_buffers[i].iProtocol))
+			continue;
+
+		    if ((proto_buffers[i].dwServiceFlags1 & XP1_IFS_HANDLES) == 0)
+			continue;
+
+		    out = WSASocket(af, type, protocol, &(proto_buffers[i]), 0, 0);
+		    break;
+		}
+	    }
+
+	    free(proto_buffers);
+	}
+    }
+
+    return out;
+}
+#endif	/* USE_WINSOCK2 */
+
 #undef socket
+#ifdef USE_WINSOCK2
+#define open_socket(a, t, p)	open_ifs_socket(a, t, p)
+#else
+#define open_socket(a, t, p)	socket(a, t, p)
+#endif
 
 int 
 rb_w32_socket(int af, int type, int protocol)
@@ -2263,7 +2361,7 @@ rb_w32_socket(int af, int type, int protocol)
 	StartSockets();
     }
     RUBY_CRITICAL({
-	s = socket(af, type, protocol);
+	s = open_socket(af, type, protocol);
 	if (s == INVALID_SOCKET) {
 	    errno = map_errno(WSAGetLastError());
 	    fd = -1;
@@ -2576,24 +2674,45 @@ gettimeofday(struct timeval *tv, struct timezone *tz)
 char *
 rb_w32_getcwd(char *buffer, int size)
 {
-    int length;
+    char *p = buffer;
     char *bp;
+    int len;
 
-#undef getcwd
-    if (getcwd(buffer, size) == NULL) {
-        return NULL;
-    }
-    length = strlen(buffer);
-    if (length >= size) {
-        return NULL;
+    len = GetCurrentDirectory(0, NULL);
+    if (!len) {
+	errno = map_errno(GetLastError());
+	return NULL;
     }
 
-    for (bp = buffer; *bp != '\0'; bp = CharNext(bp)) {
+    if (p) {
+	if (size < len) {
+	    errno = ERANGE;
+	    return NULL;
+	}
+    }
+    else {
+	p = malloc(len);
+	size = len;
+	if (!p) {
+	    errno = ENOMEM;
+	    return NULL;
+	}
+    }
+
+    if (!GetCurrentDirectory(size, p)) {
+	errno = map_errno(GetLastError());
+	if (!buffer)
+	    free(p);
+	return NULL;
+    }
+
+    for (bp = p; *bp != '\0'; bp = CharNext(bp)) {
 	if (*bp == '\\') {
 	    *bp = '/';
 	}
     }
-    return buffer;
+
+    return p;
 }
 
 int
@@ -2825,6 +2944,112 @@ isUNCRoot(const char *path)
     return 0;
 }
 
+static time_t
+filetime_to_unixtime(const FILETIME *ft)
+{
+    FILETIME loc;
+    SYSTEMTIME st;
+    struct tm tm;
+    time_t t;
+
+    if (!FileTimeToLocalFileTime(ft, &loc)) {
+	return 0;
+    }
+    if (!FileTimeToSystemTime(&loc, &st)) {
+	return 0;
+    }
+    memset(&tm, 0, sizeof(tm));
+    tm.tm_year = st.wYear - 1900;
+    tm.tm_mon = st.wMonth - 1;
+    tm.tm_mday = st.wDay;
+    tm.tm_hour = st.wHour;
+    tm.tm_min = st.wMinute;
+    tm.tm_sec = st.wSecond;
+    tm.tm_isdst = -1;
+    t = mktime(&tm);
+    return t == -1 ? 0 : t;
+}
+
+static unsigned
+fileattr_to_unixmode(DWORD attr, const char *path)
+{
+    unsigned mode = 0;
+
+    if (attr & FILE_ATTRIBUTE_READONLY) {
+	mode |= S_IREAD;
+    }
+    else {
+	mode |= S_IREAD | S_IWRITE | S_IWUSR;
+    }
+
+    if (attr & FILE_ATTRIBUTE_DIRECTORY) {
+	mode |= S_IFDIR | S_IEXEC;
+    }
+    else {
+	mode |= S_IFREG;
+    }
+
+    if (path && (mode & S_IFREG)) {
+	const char *end = path + strlen(path);
+	while (path < end) {
+	    end = CharPrev(path, end);
+	    if (*end == '.') {
+		if ((strcmpi(end, ".bat") == 0) ||
+		    (strcmpi(end, ".cmd") == 0) ||
+		    (strcmpi(end, ".com") == 0) ||
+		    (strcmpi(end, ".exe") == 0)) {
+		    mode |= S_IEXEC;
+		}
+		break;
+	    }
+	}
+    }
+
+    mode |= (mode & 0700) >> 3;
+    mode |= (mode & 0700) >> 6;
+
+    return mode;
+}
+
+static int
+winnt_stat(const char *path, struct stat *st)
+{
+    HANDLE h;
+    WIN32_FIND_DATA wfd;
+
+    memset(st, 0, sizeof(struct stat));
+    st->st_nlink = 1;
+
+    if (_mbspbrk(path, "?*")) {
+	errno = ENOENT;
+	return -1;
+    }
+    h = FindFirstFile(path, &wfd);
+    if (h != INVALID_HANDLE_VALUE) {
+	FindClose(h);
+	st->st_mode  = fileattr_to_unixmode(wfd.dwFileAttributes, path);
+	st->st_atime = filetime_to_unixtime(&wfd.ftLastAccessTime);
+	st->st_mtime = filetime_to_unixtime(&wfd.ftLastWriteTime);
+	st->st_ctime = filetime_to_unixtime(&wfd.ftCreationTime);
+	st->st_size  = wfd.nFileSizeLow; /* TODO: 64bit support */
+    }
+    else {
+	// If runtime stat(2) is called for network shares, it fails on WinNT.
+	// Because GetDriveType returns 1 for network shares. (Win98 returns 4)
+	DWORD attr = GetFileAttributes(path);
+	if (attr == -1) {
+	    errno = map_errno(GetLastError());
+	    return -1;
+	}
+	st->st_mode  = fileattr_to_unixmode(attr, path);
+    }
+
+    st->st_dev = st->st_rdev = (isalpha(path[0]) && path[1] == ':') ?
+	toupper(path[0]) - 'A' : _getdrive() - 1;
+
+    return 0;
+}
+
 int
 rb_w32_stat(const char *path, struct stat *st)
 {
@@ -2860,7 +3085,7 @@ rb_w32_stat(const char *path, struct stat *st)
     } else if (*end == '\\' || (buf1 + 1 == end && *end == ':'))
 	strcat(buf1, ".");
 
-    ret = stat(buf1, st);
+    ret = IsWinNT() ? winnt_stat(buf1, st) : stat(buf1, st);
     if (ret == 0) {
 	st->st_mode &= ~(S_IWGRP | S_IWOTH);
     }
@@ -3331,6 +3556,30 @@ rb_w32_close(int fd)
     return 0;
 }
 
+#undef read
+size_t
+rb_w32_read(int fd, void *buf, size_t size)
+{
+    SOCKET sock = TO_SOCKET(fd);
+
+    if (!is_socket(sock))
+	return read(fd, buf, size);
+    else
+	return rb_w32_recv(fd, buf, size, 0);
+}
+
+#undef write
+size_t
+rb_w32_write(int fd, const void *buf, size_t size)
+{
+    SOCKET sock = TO_SOCKET(fd);
+
+    if (!is_socket(sock))
+	return write(fd, buf, size);
+    else
+	return rb_w32_send(fd, buf, size, 0);
+}
+
 static int
 unixtime_to_filetime(time_t time, FILETIME *ft)
 {
@@ -3377,6 +3626,9 @@ rb_w32_utime(const char *path, struct utimbuf *times)
     }
 
     RUBY_CRITICAL({
+	const DWORD attr = GetFileAttributes(path);
+	if (attr != (DWORD)-1 && (attr & FILE_ATTRIBUTE_READONLY))
+	    SetFileAttributes(path, attr & ~FILE_ATTRIBUTE_READONLY);
 	hFile = CreateFile(path, GENERIC_WRITE, 0, 0, OPEN_EXISTING,
 			   IsWin95() ? 0 : FILE_FLAG_BACKUP_SEMANTICS, 0);
 	if (hFile == INVALID_HANDLE_VALUE) {
@@ -3390,6 +3642,8 @@ rb_w32_utime(const char *path, struct utimbuf *times)
 	    }
 	    CloseHandle(hFile);
 	}
+	if (attr != (DWORD)-1 && (attr & FILE_ATTRIBUTE_READONLY))
+	    SetFileAttributes(path, attr);
     });
 
     return ret;
